package org.processmining.dataawarereplayer.precision.measure;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Set;

import org.deckfour.xes.extension.std.XOrganizationalExtension;
import org.processmining.dataawarereplayer.precision.PrecisionConfig;
import org.processmining.dataawarereplayer.precision.projection.ProcessProjector;
import org.processmining.dataawarereplayer.precision.projection.ProjectedEvent;
import org.processmining.dataawarereplayer.precision.state.PrecisionState;
import org.processmining.dataawarereplayer.precision.state.StateFactory;
import org.processmining.dataawarereplayer.precision.state.StateTransition;
import org.processmining.dataawarereplayer.precision.state.StateTransitionResourceAndActivityImpl;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.semantics.petrinet.EfficientPetrinetSemantics;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataState;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateHelper;
import org.processmining.plugins.balancedconformance.dataflow.util.ExpressionEvaluator;

import com.google.common.collect.ImmutableSet;

public final class ResourcePrecisionMeasureImpl extends AbstractDPNPrecisionMeasure {

	private final static class TransitionAndResourceContinuationImpl implements ProjectedEvent {

		private final Transition nextTransition;
		private final String resource;

		private TransitionAndResourceContinuationImpl(Transition nextTransition, String resource) {
			this.nextTransition = nextTransition;
			this.resource = resource;
		}

		public Set<String> getAttributesWritten() {
			return ImmutableSet.of(XOrganizationalExtension.KEY_RESOURCE);
		}

		public Object getAttributeValue(String varName) {
			if (varName.equals(XOrganizationalExtension.KEY_RESOURCE)) {
				return resource;
			}
			return null;
		}

		public Transition getTransition() {
			return nextTransition;
		}

		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((nextTransition == null) ? 0 : nextTransition.hashCode());
			result = prime * result + ((resource == null) ? 0 : resource.hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (!(obj instanceof TransitionAndResourceContinuationImpl))
				return false;
			TransitionAndResourceContinuationImpl other = (TransitionAndResourceContinuationImpl) obj;
			if (nextTransition == null) {
				if (other.nextTransition != null)
					return false;
			} else if (!nextTransition.equals(other.nextTransition))
				return false;
			if (resource == null) {
				if (other.resource != null)
					return false;
			} else if (!resource.equals(other.resource))
				return false;
			return true;
		}

	}

	public ResourcePrecisionMeasureImpl(PrecisionConfig config, ProcessProjector processProjector,
			StateFactory stateFactory, DataPetriNet net) {
		super(config, processProjector, stateFactory, net);
	}

	protected StateTransition newStateTransition(ProjectedEvent event) {
		// Query resource directly as it is not considered in the 'state'
		Object resource = event.getAttributeValue(XOrganizationalExtension.KEY_RESOURCE);
		if (resource != null) {
			return new StateTransitionResourceAndActivityImpl(event.getTransition(), (String) resource);
		} else {
			return new StateTransitionResourceAndActivityImpl(event.getTransition(), "NONE");
		}
	}

	protected Collection<StateTransition> recordPossibilities(DataState variableAssignment, PrecisionState state,
			EfficientPetrinetSemantics petriNetSemantics) {
		Collection<StateTransition> continuations = new ArrayList<>();
		for (Transition transition : transitions) {
			if (petriNetSemantics.isEnabled(transition)) {
				Set<String> possibleExecutors = config.getExecutors().get(transition);
				if (transition instanceof PNWDTransition
						&& ((PNWDTransition) transition).getGuardExpression() != null) {
					GuardExpression guardExpression = ((PNWDTransition) transition).getGuardExpression();
					boolean allVariablesAssigned = ExpressionEvaluator.checkAllVariableNotNull(guardExpression,
							variableAssignment);
					if (allVariablesAssigned) {
						// Guard fulfilled can execute with every possible resource
						for (String executor : possibleExecutors) {
							continuations.add(newStateTransition(
									new TransitionAndResourceContinuationImpl(transition, executor)));
						}
					} else {
						if (expressionContainsResourceAssignmentMap.contains(transition)) {
							// Filter list of executors by guard
							DataState clonedAssignment = variableAssignment.clone(false);
							for (String executor : possibleExecutors) {
								// Try guard for each possible executor
								DataStateHelper.storeValue(clonedAssignment,
										DataStateHelper.addPrime(XOrganizationalExtension.KEY_RESOURCE), executor,
										String.class, processProjector.getStringDiscretizer());
								if (ExpressionEvaluator.isTrue(guardExpression, clonedAssignment,
										processProjector.getAttributeType(), processProjector.getStringDiscretizer())) {
									// Executor is allowed according to guard
									continuations.add(newStateTransition(
											new TransitionAndResourceContinuationImpl(transition, executor)));
								}
							}
						} else {
							// Assume write operations are OK, as this guard has nothing to do with resources
							for (String executor : possibleExecutors) {
								continuations.add(newStateTransition(
										new TransitionAndResourceContinuationImpl(transition, executor)));
							}
						}
					}
				} else {
					// No guard can execute with every possible resource
					for (String executor : possibleExecutors) {
						continuations.add(
								newStateTransition(new TransitionAndResourceContinuationImpl(transition, executor)));
					}
				}
			}
		}
		return continuations;
	}

}