package org.processmining.dataawarereplayer.precision.visualizer;

import java.awt.Color;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.processmining.dataawarereplayer.precision.projection.ProjectedEvent;
import org.processmining.dataawarereplayer.precision.state.PrecisionState;
import org.processmining.dataawarereplayer.precision.state.StateTransition;
import org.processmining.dataawarereplayer.precision.state.TraceStatePrefixImpl;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Event;

public class PrecisionTrace extends ProMTraceView.AbstractTrace<Event> {

	private static final String LABEL_POSSIBLE_BEHAVIOR = "Possible Behavior";
	private static final String LABEL_OBSERVED_BEHAVIOR = "Observed Behavior";

	private static final Color IMPRECISE_COLOR = new Color(252, 141, 98);
	private static final Color PRECISE_COLOR = new Color(102, 194, 165);

	private final long frequency;
	private final List<Event> sequence;

	public PrecisionTrace(PrecisionState state, Collection<StateTransition> observerdCont,
			Collection<StateTransition> possibleCont, long frequency) {
		this.frequency = frequency;
		this.sequence = getSequence(state, observerdCont, possibleCont);
	}

	@Override
	protected List<Event> delegate() {
		return sequence;
	}

	@Override
	public String getName() {
		return MessageFormat.format("{0}x", getFrequency());
	}

	public long getFrequency() {
		return frequency;
	}

	private static List<Event> getSequence(final PrecisionState state, Collection<StateTransition> observerdCont,
			Collection<StateTransition> possibleCont) {
		List<Event> sequence = new ArrayList<>();

		if (state.getTraceState() instanceof TraceStatePrefixImpl) {
			TraceStatePrefixImpl traceState = (TraceStatePrefixImpl) state.getTraceState();

			Map<String, Object> currentVariableAssignment = new HashMap<>();

			for (Iterator<ProjectedEvent> iterator = traceState.getPrefix().iterator(); iterator.hasNext();) {

				final ProjectedEvent event = iterator.next();

				updateVariableAssignment(currentVariableAssignment, event);

				if (iterator.hasNext()) {
					sequence.add(new ProMTraceView.AbstractEvent() {

						public String getLabel() {
							return event.getTransition().getLabel();
						}

						public Color getWedgeColor() {
							return Color.WHITE;
						}

					});
				} else {
					addLastEvent(sequence, currentVariableAssignment, event);
				}
			}

		} else {
			// Just use the toString representation of our state
			final String stateLabel = state.toString();
			sequence.add(new ProMTraceView.AbstractEvent() {

				public String getLabel() {					
					return stateLabel;
				}

				public Color getWedgeColor() {
					return Color.WHITE;
				}

			});
		}
		addStateTransitions(sequence, observerdCont, possibleCont);
		return sequence;
	}

	private static void updateVariableAssignment(Map<String, Object> currentVariableAssignment,
			final ProjectedEvent event) {
		for (Iterator<String> attrIter = event.getAttributesWritten().iterator(); attrIter.hasNext();) {
			String attr = attrIter.next();
			currentVariableAssignment.put(attr, event.getAttributeValue(attr));
		}
	}

	private static void addStateTransitions(List<Event> sequence, Collection<StateTransition> observerdCont,
			final Collection<StateTransition> possibleCont) {
		if (observerdCont.equals(possibleCont)) {
			final String eventLabel = observerdCont.toString();
			sequence.add(new ProMTraceView.AbstractEvent() {

				public String getLabel() {					
					return eventLabel;
				}

				public Color getWedgeColor() {
					return PRECISE_COLOR;
				}

				public String getTopLabel() {
					return LABEL_OBSERVED_BEHAVIOR;
				}

				public String getBottomLabel() {
					return LABEL_POSSIBLE_BEHAVIOR;
				}

			});
		} else {
			final String upperLabel = observerdCont.toString();
			final String lowerLabel = possibleCont.toString();
			final boolean isPrecise = observerdCont.equals(possibleCont);
			sequence.add(new ProMTraceView.AbstractSplittedEvent() {

				public String getUpperLabel() {					
					return upperLabel;
				}

				public Color getUpperColor() {					
					if (isPrecise) {
						return PRECISE_COLOR;
					} else {
						return IMPRECISE_COLOR;
					}
				}

				public String getLowerLabel() {					
					return lowerLabel;
				}

				public Color getLowerColor() {
					return PRECISE_COLOR;
				}

				public String getTopLabel() {
					return LABEL_OBSERVED_BEHAVIOR;
				}

				public String getBottomLabel() {
					return LABEL_POSSIBLE_BEHAVIOR;
				}

			});
		}
	}

	private static void addLastEvent(List<Event> sequence, Map<String, Object> currentVariableAssignment,
			final ProjectedEvent event) {
		StringBuilder sb = new StringBuilder(currentVariableAssignment.size() * 20 + 2);
		sb.append("[");
		for (Iterator<Entry<String, Object>> attrIter = currentVariableAssignment.entrySet().iterator(); attrIter
				.hasNext();) {
			Entry<String, Object> attr = attrIter.next();
			sb.append(attr.getKey()); 
			sb.append(" = ");
			sb.append(attr.getValue());
			if (attrIter.hasNext()) {
				sb.append(", ");
			}
		}
		sb.append("]");
		final String variableWritings = sb.toString();
		sequence.add(new ProMTraceView.AbstractSplittedEvent() {

			public String getUpperLabel() {
				return event.getTransition().getLabel();
			}

			public String getLowerLabel() {
				return variableWritings;
			}

			public Color getWedgeColor() {
				return Color.WHITE;
			}

			public Color getUpperColor() {
				return getWedgeColor();
			}

			public Color getLowerColor() {
				return getWedgeColor();
			}

		});
	}

}