/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.config;

import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import nl.tue.astar.AStarThread.ASynchronousMoveSorting;
import nl.tue.astar.AStarThread.QueueingModel;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.ui.ImprovedEvClassLogMappingUI;
import org.processmining.log.utils.XUtils;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.DataConformance.framework.ReplayableActivity;
import org.processmining.plugins.DataConformance.framework.VariableMatchCost;
import org.processmining.plugins.DataConformance.framework.VariableMatchCosts;
import org.processmining.plugins.balancedconformance.controlflow.adapter.AlignmentAdapter.PriorityQueueFactory;
import org.processmining.plugins.balancedconformance.controlflow.adapter.SearchMethod;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapter.ILPSolver;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;

/**
 * Configuration for the {@link DataAlignmentTraceProcessor}. Most configuration
 * parameters stored in this class are set to reasonable defaults. Use the
 * constructor that requires all mandatory parameters
 * {@link #BalancedProcessorConfiguration(Marking, Marking[], TransEvClassMapping, Map, Map, Map, VariableMatchCosts, Map)}
 * . Otherwise you need to use the following methods to make sure that these
 * mandatory parameters are set according to your needs:
 * <p>
 * {@link #setActivityMapping(org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping)}
 * ,
 * {@link #setCostBasedCompleteParam(org.processmining.plugins.petrinet.replayer.algorithms.costbasedcomplete.CostBasedCompleteParam)}
 * , {@link #setLowerBounds(java.util.Map)},
 * {@link #setUpperBounds(java.util.Map)},
 * {@link #setVariableCost(org.processmining.plugins.DataConformance.framework.VariableMatchCosts)}
 * , {@link #setVariableMapping(java.util.Map)}.
 * 
 * @author F. Mannhardt
 * 
 */
/**
 * @author fmannhar
 * 
 */
public class BalancedProcessorConfiguration extends DataConformancePlusConfiguration {

	private static final int VARIABLE_BOUNDS_PRECISION = 1000;

	public enum UnassignedMode {
		NULL, DEFAULT, FREE
	}

	public enum DataStateStorageHandlerType {
		CAS, RWLOCK, MEMORY_OPTIMIZED, PRIMITIVE_NOLOCK, MEMORY_OPTIMIZED_NOLOCK, JAVA_NOLOCK
	}

	public enum ControlFlowStorageHandlerType {
		MEMORY_EFFICIENT_LOCKING, CPU_EFFICIENT_LOCKING, MEMORY_EFFICIENT, CPU_EFFICIENT,
	}

	// A* settings
	private SearchMethod searchMethod = SearchMethod.ASTAR_GRAPH;
	private ControlFlowStorageHandlerType controlFlowStorageHandlerType = ControlFlowStorageHandlerType.MEMORY_EFFICIENT_LOCKING;
	private QueueingModel queueingModel = QueueingModel.DEPTHFIRST;
	private ASynchronousMoveSorting sorting = ASynchronousMoveSorting.LOGMOVEFIRST;

	/**
	 * assume at most 100 tau transitions
	 */
	private int costDelta = 100;

	/**
	 * maximum number of queued states for one trace
	 */
	private int maxQueuedStates = Integer.MAX_VALUE;

	/**
	 * maximum time spend for one trace in seconds
	 */
	private long timeLimitPerTrace = Long.MAX_VALUE;

	private double maxCostFactor;

	// Data alignment (ILP) settings
	private boolean isUseOptimizations = true;
	private ILPSolver ilpSolver = ILPSolver.ILP_LPSOLVE;
	private DataStateStorageHandlerType dataStateStorageHandlerType = DataStateStorageHandlerType.CAS;
	private UnassignedMode variablesUnassignedMode = UnassignedMode.FREE;
	private boolean isUsePartialDataAlignments = true;

	// Experimental settings
	private boolean includeVirtualVariablesInTrace = false;

	public static BalancedProcessorConfiguration newDefaultInstance(PetrinetGraph net, Marking initialMarking,
			Marking[] finalMarkings, XLog log, XEventClassifier classifier, int defaultMoveOnModelCost,
			int defaultMoveOnLogCost, int defaultMissingWriteOpCost, int defaultIncorrectWriteOpCost) {
		return newDefaultInstance(net, initialMarking, finalMarkings, log, classifier, defaultMoveOnModelCost,
				defaultMoveOnLogCost, defaultMissingWriteOpCost, defaultIncorrectWriteOpCost, false);
	}

	/**
	 * Creates a {@link BalancedProcessorConfiguration} with reasonable default
	 * settings, i.e. the name of attributes in the event log has to exactly
	 * match variable names in the DPN-net, and the name of the event classes
	 * has to exactly match the transition names in the DPN-net.
	 * 
	 * @param net
	 * @param initialMarking
	 * @param finalMarkings
	 * @param log
	 * @param classifier
	 * @param defaultMoveOnModelCost
	 * @param defaultMoveOnLogCost
	 * @param defaultMissingWriteOpCost
	 * @param defaultIncorrectWriteOpCost
	 * @return
	 */
	public static BalancedProcessorConfiguration newDefaultInstance(PetrinetGraph net, Marking initialMarking,
			Marking[] finalMarkings, XLog log, XEventClassifier classifier, int defaultMoveOnModelCost,
			int defaultMoveOnLogCost, int defaultMissingWriteOpCost, int defaultIncorrectWriteOpCost,
			boolean usePartialorder) {

		XEventClasses eventClasses = XUtils.createEventClasses(classifier, log);
		Set<String> attributeNames = getAttributeNames(log, eventClasses.getClassifier());

		return newDefaultInstance(net, initialMarking, finalMarkings, eventClasses, attributeNames,
				defaultMoveOnModelCost, defaultMoveOnLogCost, defaultMissingWriteOpCost, defaultIncorrectWriteOpCost,
				usePartialorder);
	}

	/**
	 * Creates a {@link BalancedProcessorConfiguration} with reasonable default
	 * settings, i.e. the name of attributes in the event log has to exactly
	 * match variable names in the DPN-net, and the name of the event classes
	 * has to exactly match the transition names in the DPN-net.
	 * 
	 * @param net
	 * @param initialMarking
	 * @param finalMarkings
	 * @param eventClasses
	 * @param attributeNames
	 * @param defaultMoveOnModelCost
	 * @param defaultMoveOnLogCost
	 * @param defaultMissingWriteOpCost
	 * @param defaultIncorrectWriteOpCost
	 * @return
	 */
	public static BalancedProcessorConfiguration newDefaultInstance(PetrinetGraph net, Marking initialMarking,
			Marking[] finalMarkings, XEventClasses eventClasses, Set<String> attributeNames,
			int defaultMoveOnModelCost, int defaultMoveOnLogCost, int defaultMissingWriteOpCost,
			int defaultIncorrectWriteOpCost) {
		return newDefaultInstance(net, initialMarking, finalMarkings, eventClasses, attributeNames,
				defaultMoveOnModelCost, defaultMoveOnLogCost, defaultMissingWriteOpCost, defaultIncorrectWriteOpCost,
				false);
	}

	public static BalancedProcessorConfiguration newDefaultInstance(PetrinetGraph net, Marking initialMarking,
			Marking[] finalMarkings, XEventClasses eventClasses, Set<String> attributeNames,
			int defaultMoveOnModelCost, int defaultMoveOnLogCost, int defaultMissingWriteOpCost,
			int defaultIncorrectWriteOpCost, boolean usePartialorder) {

		Set<String> variableNames;
		if (net instanceof DataPetriNet) {
			variableNames = PetriNetWithData.getAllVariableNames((DataPetriNet) net);
		} else {
			variableNames = ImmutableSet.of();
		}

		TransEvClassMapping activityMapping = createDefaultMappingTransitionsToEventClasses(net,
				eventClasses.getClassifier(), eventClasses);
		Map<XEventClass, Integer> mapEvClass2Cost = createDefaultLogMoveCost(eventClasses, defaultMoveOnLogCost);
		Map<Transition, Integer> mapTrans2Cost = createDefaultModelMoveCost(net, defaultMoveOnModelCost);

		Map<String, String> variableMapping;
		if (!variableNames.isEmpty()) {
			variableMapping = createDefaultVariableMapping(variableNames, attributeNames);
		} else {
			variableMapping = ImmutableMap.of();
		}

		VariableMatchCosts variableCost;
		if (net instanceof DataPetriNet) {
			variableCost = createDefaultVariableCost((DataPetriNet) net, variableNames, defaultMissingWriteOpCost,
					defaultIncorrectWriteOpCost);
		} else {
			variableCost = new VariableMatchCosts(0, 0, ImmutableSet.<ReplayableActivity>of(),
					ImmutableSet.<String>of());
		}

		return new BalancedProcessorConfiguration(initialMarking, finalMarkings, activityMapping, mapEvClass2Cost,
				mapTrans2Cost, variableMapping, variableCost, usePartialorder);
	}

	private static Set<String> getAttributeNames(XLog log, XEventClassifier classifier) {
		Set<String> attributeNames;
		if (log.getInfo(classifier) != null) {
			attributeNames = ImmutableSet.copyOf(log.getInfo(classifier).getEventAttributeInfo().getAttributeKeys());
		} else {
			attributeNames = XUtils.getEventAttributeKeys(log);
		}
		return attributeNames;
	}

	@Deprecated
	public static Map<String, String> createDefaultVariableMapping(XLogInfo logInfo, Set<String> variableNames) {
		Map<String, String> mapping = new HashMap<>();
		for (String attributeKey : logInfo.getEventAttributeInfo().getAttributeKeys()) {
			if (variableNames.contains(attributeKey)) {
				mapping.put(attributeKey, attributeKey);
			}
		}
		return mapping;
	}

	public static Map<String, String> createDefaultVariableMapping(Set<String> variableNames, Set<String> attributeNames) {
		Map<String, String> mapping = new HashMap<>();
		for (String attributeKey : attributeNames) {
			if (variableNames.contains(attributeKey)) {
				mapping.put(attributeKey, attributeKey);
			}
		}
		return mapping;
	}

	public static VariableMatchCosts createDefaultVariableCost(DataPetriNet net, Set<String> variableSet,
			int defaultMissingWriteOpCost, int defaultIncorrectWriteOpCost) {
		Set<ReplayableActivity> activitySet = new HashSet<>();
		for (final Transition t : net.getTransitions()) {
			activitySet.add(new ReplayableActivity() {

				public String getLabel() {
					return t.getLabel();
				}
			});
		}
		VariableMatchCosts variableMatchCost = new VariableMatchCosts(defaultIncorrectWriteOpCost,
				defaultMissingWriteOpCost, activitySet, variableSet);
		return variableMatchCost;
	}

	public static VariableMatchCosts createCustomVariableCost(DataPetriNet net, List<VariableMatchCost> variableCost) {
		Set<ReplayableActivity> activitySet = new HashSet<>();
		for (final Transition t : net.getTransitions()) {
			activitySet.add(new ReplayableActivity() {

				public String getLabel() {
					return t.getLabel();
				}
			});
		}
		Set<String> variableNames = PetriNetWithData.getAllVariableNames(net);
		return new VariableMatchCosts(variableCost, activitySet, variableNames);
	}

	public static Map<Transition, Integer> createDefaultModelMoveCost(PetrinetGraph net, int defaultMoveOnModelCost) {
		Map<Transition, Integer> mapTrans2Cost = new HashMap<>();
		for (Transition t : net.getTransitions()) {
			if (t.isInvisible()) {
				mapTrans2Cost.put(t, 0);
			} else {
				mapTrans2Cost.put(t, defaultMoveOnModelCost);
			}
		}
		return mapTrans2Cost;
	}

	public static Map<XEventClass, Integer> createDefaultLogMoveCost(XLogInfo info, int defaultMoveOnLogCost) {
		return createDefaultLogMoveCost(info.getEventClasses(), defaultMoveOnLogCost);
	}

	public static Map<XEventClass, Integer> createDefaultLogMoveCost(XEventClasses eventClasses,
			int defaultMoveOnLogCost) {
		Map<XEventClass, Integer> mapEvClass2Cost = new HashMap<>();
		for (XEventClass eventClass : eventClasses.getClasses()) {
			mapEvClass2Cost.put(eventClass, defaultMoveOnLogCost);
		}
		mapEvClass2Cost.put(ImprovedEvClassLogMappingUI.DUMMY, 0);
		return mapEvClass2Cost;
	}

	public static TransEvClassMapping createDefaultMappingTransitionsToEventClasses(PetrinetGraph net,
			XEventClassifier classifier, XLogInfo info) {
		return createDefaultMappingTransitionsToEventClasses(net, classifier, info.getEventClasses());
	}

	public static TransEvClassMapping createDefaultMappingTransitionsToEventClasses(PetrinetGraph net,
			XEventClassifier classifier, XEventClasses eventClasses) {
		TransEvClassMapping activityMapping = new TransEvClassMapping(classifier, ImprovedEvClassLogMappingUI.DUMMY);
		for (Transition t : net.getTransitions()) {
			for (XEventClass eventClass : eventClasses.getClasses()) {
				if (eventClass.getId().equalsIgnoreCase(t.getLabel())) {
					activityMapping.put(t, eventClass);
				}
			}
		}
		for (Transition t : net.getTransitions()) {
			if (!activityMapping.containsKey(t)) {
				activityMapping.put(t, activityMapping.getDummyEventClass());
			}
		}
		return activityMapping;
	}

	public static void autoGuessBounds(BalancedProcessorConfiguration config, DataPetriNet currentNet, XLog log) {

		Map<String, Object> upperBounds = new HashMap<>();
		Map<String, Object> lowerBounds = new HashMap<>();

		Collection<DataElement> variables = currentNet.getVariables();

		if (!variables.isEmpty()) {

			Map<DataElement, Number> minValueMap = new HashMap<>();
			Map<DataElement, Number> maxValueMap = new HashMap<>();

			for (XTrace trace : log) {
				for (XEvent event : trace) {

					XAttributeMap attributes = event.getAttributes();
					for (DataElement elem : variables) {
						String attributeName = config.getVariableMapping().get(elem.getVarName());
						XAttribute attribute = attributes.get(attributeName);
						if (attribute != null) {
							updateBounds(minValueMap, maxValueMap, attribute, elem);
						}
					}
				}
			}

			for (DataElement elem : variables) {

				Object row[] = new Object[4];
				row[0] = elem.getVarName();
				row[1] = elem.getType();

				Number minValue = determineMinValue(elem, minValueMap);
				Number maxValue = determineMaxValue(elem, maxValueMap);

				if (minValue != null && minValue.equals(maxValue)) {
					maxValue = minValue.longValue() + VARIABLE_BOUNDS_PRECISION;
				}

				if (minValue != null) {
					if (elem.getType() == Date.class) {
						// default allow all dates
						lowerBounds.put(elem.getVarName(), new Date(0));
					} else {
						lowerBounds.put(elem.getVarName(), minValue);
					}
				}

				if (maxValue != null) {
					if (elem.getType() == Date.class) {
						upperBounds.put(elem.getVarName(), new Date(maxValue.longValue()));
					} else {
						upperBounds.put(elem.getVarName(), maxValue);
					}
				}
			}
		}

		config.setUpperBounds(upperBounds);
		config.setLowerBounds(lowerBounds);
	}

	private static void updateBounds(Map<DataElement, Number> minValueMap, Map<DataElement, Number> maxValueMap,
			XAttribute attribute, DataElement elem) {
		Number minValue = minValueMap.get(elem);
		Number maxValue = maxValueMap.get(elem);

		if (attribute instanceof XAttributeDiscrete) {
			XAttributeDiscrete dAttr = (XAttributeDiscrete) attribute;
			long value = dAttr.getValue();
			minValueMap.put(elem, minValue == null ? value : Math.min(minValue.longValue(), value));
			maxValueMap.put(elem, maxValue == null ? value : Math.max(maxValue.longValue(), value));
		} else if (attribute instanceof XAttributeContinuous) {
			XAttributeContinuous cAttr = (XAttributeContinuous) attribute;
			double value = cAttr.getValue();
			minValueMap.put(elem, minValue == null ? value : Math.min(minValue.doubleValue(), value));
			maxValueMap.put(elem, maxValue == null ? value : Math.max(maxValue.doubleValue(), value));
		} else if (attribute instanceof XAttributeTimestamp) {
			XAttributeTimestamp tAttr = (XAttributeTimestamp) attribute;
			long value = tAttr.getValueMillis();
			minValueMap.put(elem, minValue == null ? value : Math.min(minValue.longValue(), value));
			maxValueMap.put(elem, maxValue == null ? value : Math.max(maxValue.longValue(), value));
		}
	}

	private static Number determineMaxValue(DataElement elem, Map<DataElement, Number> maxValueMap) {
		if (elem.getMaxValue() != null && elem.getMaxValue() instanceof Number) {
			return (Number) elem.getMaxValue();
		} else {
			Number maxValue = maxValueMap.get(elem);
			if (maxValue != null) {
				double roundedValue = Math.ceil((maxValue.doubleValue() / VARIABLE_BOUNDS_PRECISION))
						* VARIABLE_BOUNDS_PRECISION;
				if (elem.getType() == Long.class) {
					return new Long((int) roundedValue);
				} else if (elem.getType() == Integer.class) {
					return new Integer((int) roundedValue);
				} else if (elem.getType() == Double.class) {
					return new Double(roundedValue);
				} else if (elem.getType() == Float.class) {
					return new Float(roundedValue);
				} else if (elem.getType() == Date.class) {
					return new Long((long) roundedValue);
				} else {
					return null;
				}
			} else {
				return null;
			}
		}
	}

	private static Number determineMinValue(DataElement elem, Map<DataElement, Number> minValueMap) {
		if (elem.getMinValue() != null && elem.getMinValue() instanceof Number) {
			return (Number) elem.getMinValue();
		} else {
			Number minValue = minValueMap.get(elem);
			if (minValue != null) {
				double roundedValue = Math.floor((minValue.doubleValue() / VARIABLE_BOUNDS_PRECISION))
						* VARIABLE_BOUNDS_PRECISION;
				if (elem.getType() == Long.class) {
					return new Long((long) roundedValue);
				} else if (elem.getType() == Integer.class) {
					return new Integer((int) roundedValue);
				} else if (elem.getType() == Double.class) {
					return new Double(roundedValue);
				} else if (elem.getType() == Float.class) {
					return new Float(roundedValue);
				} else if (elem.getType() == Date.class) {
					return new Long((long) roundedValue);
				} else {
					return null;
				}
			} else {
				return null;
			}
		}
	}

	/**
	 * Creates an {@link BalancedProcessorConfiguration} with reasonable
	 * defaults, you still need to set the mandatory parameters (see other
	 * constructor) manually.
	 */
	public BalancedProcessorConfiguration() {
		super();
	}

	/**
	 * Creates a new configuration with the supplied mandatory parameters. All
	 * other configuration options are set to reasonable defaults.
	 * 
	 * @param initialMarking
	 * @param finalMarkings
	 * @param activityMapping
	 *            Mapping of Transition to Event Class
	 * @param mapEvClass2Cost
	 * @param mapTrans2Cost
	 * @param costBasedCompleteParam
	 *            Configuration of the PNReplayer
	 * @param variableMapping
	 *            Mapping from variable name in DPN-net to variable name in
	 *            event log
	 * @param variableCost
	 *            Mapping from variable name in DPN-net to costss
	 */
	public BalancedProcessorConfiguration(Marking initialMarking, Marking[] finalMarkings,
			TransEvClassMapping activityMapping, Map<XEventClass, Integer> mapEvClass2Cost,
			Map<Transition, Integer> mapTrans2Cost, Map<String, String> variableMapping,
			VariableMatchCosts variableCost, boolean usePartialOrder) {
		super(initialMarking, finalMarkings, activityMapping, mapEvClass2Cost, mapTrans2Cost, variableMapping,
				variableCost, usePartialOrder);
	}

	/**
	 * Copy constructor that takes its parameters from the supplied
	 * configuration.
	 * 
	 * @param config
	 *            to copy (shallow-clone) from
	 */
	public BalancedProcessorConfiguration(DataConformancePlusConfiguration config) {
		super(config);
	}

	/**
	 * Copy constructor that takes its parameters from the supplied
	 * configuration.
	 * 
	 * @param config
	 *            to copy (shallow-clone) from
	 */
	public BalancedProcessorConfiguration(BalancedProcessorConfiguration config) {
		super(config);
		searchMethod = config.getSearchMethod();
		controlFlowStorageHandlerType = config.getControlFlowStorageHandler();
		queueingModel = config.getQueueingModel();
		maxQueuedStates = config.getMaxQueuedStates();
		costDelta = config.getCostDelta();
		sorting = config.getSorting();
		isUseOptimizations = config.isUseOptimizations();
		ilpSolver = config.getIlpSolver();
		dataStateStorageHandlerType = config.getDataStateStorageHandler();
		variablesUnassignedMode = config.getVariablesUnassignedMode();
		isUsePartialDataAlignments = config.isUsePartialDataAlignments();
		includeVirtualVariablesInTrace = config.isIncludeVirtualVariablesInTrace();
		maxCostFactor = config.getMaxCostFactor();
	}

	public boolean isUseOptimizations() {
		return isUseOptimizations;
	}

	/**
	 * @param isUseOptimizations
	 *            whether to use an optimization that avoids calculating ILP
	 *            problem unless required
	 */
	public void setUseOptimizations(boolean isUseOptimizations) {
		this.isUseOptimizations = isUseOptimizations;
	}

	public ASynchronousMoveSorting getSorting() {
		return sorting;
	}

	public void setSorting(ASynchronousMoveSorting sorting) {
		this.sorting = sorting;
	}

	/**
	 * @param ilpSolver
	 *            which {@link ILPSolver} to use
	 */
	public void setIlpSolver(ILPSolver ilpSolver) {
		this.ilpSolver = ilpSolver;
	}

	public ILPSolver getIlpSolver() {
		return ilpSolver;
	}

	public QueueingModel getQueueingModel() {
		return queueingModel;
	}

	public void setQueueingModel(QueueingModel queueingModel) {
		this.queueingModel = queueingModel;
	}

	/**
	 * @param maxQueuedStates
	 *            stop the alignment computation after this number of states has
	 *            been added to the queue
	 */
	public void setMaxQueuedStates(int maxQueuedStates) {
		this.maxQueuedStates = maxQueuedStates;
	}

	public int getMaxQueuedStates() {
		return maxQueuedStates;
	}

	public ControlFlowStorageHandlerType getControlFlowStorageHandler() {
		return controlFlowStorageHandlerType;
	}

	public DataStateStorageHandlerType getDataStateStorageHandler() {
		return dataStateStorageHandlerType;
	}

	public void setControlFlowStorageHandler(ControlFlowStorageHandlerType controlFlowStorageHandlerType) {
		this.controlFlowStorageHandlerType = controlFlowStorageHandlerType;
	}

	public void setDataStateStorageHandler(DataStateStorageHandlerType dataStateStorageHandlerType) {
		this.dataStateStorageHandlerType = dataStateStorageHandlerType;
	}

	public boolean isUsePartialDataAlignments() {
		return isUsePartialDataAlignments;
	}

	/**
	 * @param isUsePartialDataAlignments
	 *            when enabled the "old" staged method of computing data
	 *            alignments (BPM 2013 paper) is used
	 */
	public void setUsePartialDataAlignments(boolean isUsePartialDataAlignments) {
		this.isUsePartialDataAlignments = isUsePartialDataAlignments;
	}

	public int getCostDelta() {
		return costDelta;
	}

	public void setCostDelta(int costDelta) {
		this.costDelta = costDelta;
	}

	public SearchMethod getSearchMethod() {
		return searchMethod;
	}

	public void setSearchMethod(SearchMethod searchMethod) {
		this.searchMethod = searchMethod;
	}

	public UnassignedMode getVariablesUnassignedMode() {
		return variablesUnassignedMode;
	}

	public void setVariablesUnassignedMode(UnassignedMode variablesUnassignedMode) {
		this.variablesUnassignedMode = variablesUnassignedMode;
	}

	public boolean hasUnmappedActivity(DataPetriNet currentNet) {
		for (Transition t : currentNet.getTransitions()) {
			if (!t.isInvisible() && getActivityMapping().get(t) == getActivityMapping().getDummyEventClass()) {
				return true;
			}
		}
		return false;
	}

	public boolean hasUnmappedVariable(DataPetriNet currentNet) {
		for (DataElement e : currentNet.getVariables()) {
			if (getVariableMapping().get(e.getVarName()) == null) {
				return true;
			}
		}
		return false;
	}

	public boolean getIncludeVirtualVariablesInTrace() {
		return includeVirtualVariablesInTrace;
	}

	public boolean isIncludeVirtualVariablesInTrace() {
		return includeVirtualVariablesInTrace;
	}

	public void setIncludeVirtualVariablesInTrace(boolean includeVirtualVariablesInTrace) {
		this.includeVirtualVariablesInTrace = includeVirtualVariablesInTrace;
	}

	/**
	 * @return value from 0 to 1, 0 meaning there is no limit, 1 meaning there
	 *         should be no cost
	 */
	public double getMaxCostFactor() {
		return maxCostFactor;
	}

	/**
	 * @param maxCostFactor
	 *            value from 0 to 1, 0 meaning there is no limit, 1 meaning
	 *            there should be no cost
	 */
	public void setMaxCostFactor(double maxCostFactor) {
		this.maxCostFactor = maxCostFactor;
	}

	/**
	 * Use {@link #setQueueingModel(QueueingModel)} instead
	 */
	@Deprecated
	public void setQueueingModel(PriorityQueueFactory createQueueFactory) {
	}

	public long getTimeLimitPerTrace() {
		return timeLimitPerTrace;
	}

	/**
	 * Maximum time spend for one trace in seconds
	 * 
	 * @param timeLimitPerTrace
	 */
	public void setTimeLimitPerTrace(long timeLimitPerTrace) {
		this.timeLimitPerTrace = timeLimitPerTrace;
	}

}