/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.controlflow.adapter;

import java.util.Map;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.models.graphbased.directed.petrinet.InhibitorNet;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.ResetInhibitorNet;
import org.processmining.models.graphbased.directed.petrinet.ResetNet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.astar.petrinet.impl.PHead;
import org.processmining.plugins.balancedconformance.cache.DataViewBuilder;
import org.processmining.plugins.balancedconformance.controlflow.override.DHead;
import org.processmining.plugins.balancedconformance.controlflow.override.DataAwareThread;
import org.processmining.plugins.balancedconformance.controlflow.override.XPDelegate;
import org.processmining.plugins.balancedconformance.controlflow.override.queues.duplicates.BFDataDuplicatePriorityQueue;
import org.processmining.plugins.balancedconformance.controlflow.override.queues.duplicates.DFDataDuplicatePriorityQueue;
import org.processmining.plugins.balancedconformance.controlflow.override.queues.duplicates.RandomFirstDuplicatePriorityQueue;
import org.processmining.plugins.balancedconformance.controlflow.resultbuilder.AlignmentResultBuilder;
import org.processmining.plugins.balancedconformance.controlflow.resultbuilder.AlignmentResultBuilderImpl;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapter;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;
import org.processmining.plugins.balancedconformance.observer.DataConformancePlusObserver;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;

import nl.tue.astar.AStarObserver;
import nl.tue.astar.AStarThread.QueueingModel;
import nl.tue.astar.Tail;
import nl.tue.astar.Trace;
import nl.tue.astar.impl.AbstractAStarThread;
import nl.tue.astar.impl.AbstractAStarThread.StorageHandler;
import nl.tue.astar.impl.DijkstraTail;
import nl.tue.astar.util.FastLookupPriorityQueue;

/**
 * Dijkstra-search for the best alignment taking data into account.
 * 
 * @author F. Mannhardt
 * 
 */
public class AlignmentDijkstraImpl extends AlignmentAbstractAdapter<DijkstraTail> {

	private final boolean isKeepCFSearchSpace;
	private boolean isKeepDFSearchSpace;

	public AlignmentDijkstraImpl(XLog log, DataPetriNet net, LogMapping logMapping, ControlFlowAlignmentConfig cfConfig,
			StorageHandlerFactory<DijkstraTail> storageFactory, boolean keepCFSearchSpace, boolean keepDFSearchSpace,
			DataConformancePlusObserver observer, DataAlignmentAdapter dataAlignmentAdapter) {
		super(log, net, logMapping, cfConfig, storageFactory, observer, dataAlignmentAdapter);
		this.isKeepCFSearchSpace = keepCFSearchSpace;
		this.isKeepDFSearchSpace = keepDFSearchSpace;
	}

	@Override
	protected AlignmentResultBuilder createResultBuilder(TraceMapping traceMapping,
			Map<XEvent, XEventClass> mapEv2EventClass, Trace linearTrace, AStarObserver observer) {

		StorageHandler<PHead, DijkstraTail> localStorageHandler;
		DataStateStore localDataStateStore;

		if (!isKeepCFSearchSpace) {
			localStorageHandler = storageFactory.newStorageHandler(delegate);
		} else {
			localStorageHandler = storageHandler;
		}

		if (!isKeepDFSearchSpace) {
			localDataStateStore = storageFactory.newDataStateStore();
		} else {
			localDataStateStore = dataStateStore;
		}

		traceMapping.setDataStateStore(localDataStateStore);

		DHead initialHead = new DHead(delegate, initMarking, traceMapping.getTrace());
		AbstractAStarThread<PHead, DijkstraTail> thread = new DataAwareThread<>(delegate, initialHead,
				localStorageHandler, linearTrace, maxNumOfStates, dataAdapter, logMapping, traceMapping,
				mapEv2EventClass, mapEvClass2Cost, dataViewBuilder, activateCache, useOptimizations, sorting,
				getQueueFactory(queueingModel), localDataStateStore);

		return new AlignmentResultBuilderImpl((DataAwareThread<PHead, ? extends Tail>) thread, linearTrace,
				traceMapping, canceller, delegate, mapEv2EventClass, mapEvClass2Cost, timeLimit, observer);
	}

	@Override
	protected AbstractPDelegate<DijkstraTail> getPDelegate(PetrinetGraph net, XLog log, XEventClasses classes,
			TransEvClassMapping map, Map<Transition, Integer> mapTrans2Cost, Map<XEventClass, Integer> mapEvClass2Cost,
			int delta, boolean allMarkingsAreFinal, Marking[] finalMarkings) {
		if (net instanceof ResetInhibitorNet) {
			return new XPDelegate((ResetInhibitorNet) net, log, classes, map, mapTrans2Cost, mapEvClass2Cost, delta,
					allMarkingsAreFinal, finalMarkings);
		} else if (net instanceof ResetNet) {
			return new XPDelegate((ResetNet) net, log, classes, map, mapTrans2Cost, mapEvClass2Cost, delta,
					allMarkingsAreFinal, finalMarkings);
		} else if (net instanceof InhibitorNet) {
			return new XPDelegate((InhibitorNet) net, log, classes, map, mapTrans2Cost, mapEvClass2Cost, delta,
					allMarkingsAreFinal, finalMarkings);
		} else if (net instanceof Petrinet) {
			return new XPDelegate((Petrinet) net, log, classes, map, mapTrans2Cost, mapEvClass2Cost, delta,
					allMarkingsAreFinal, finalMarkings);
		}
		throw new UnsupportedOperationException("Unkown type of Petri net " + net.getClass().getSimpleName());
	}

	private PriorityQueueFactory getQueueFactory(QueueingModel queueingModel) {
		switch (queueingModel) {
			case BREADTHFIRST :
				return new PriorityQueueFactory() {
					public FastLookupPriorityQueue newInstance(DataViewBuilder dataViewBuilder,
							AbstractPDelegate<? extends Tail> delegate) {
						return new BFDataDuplicatePriorityQueue(1024 * 2);
					}

				};

			case DEPTHFIRST :
				return new PriorityQueueFactory() {

					public FastLookupPriorityQueue newInstance(DataViewBuilder dataViewBuilder,
							AbstractPDelegate<? extends Tail> delegate) {
						return new DFDataDuplicatePriorityQueue(1024 * 2);
					}

				};

			case RANDOM :
				return new PriorityQueueFactory() {

					public FastLookupPriorityQueue newInstance(DataViewBuilder dataViewBuilder,
							AbstractPDelegate<? extends Tail> delegate) {
						return new RandomFirstDuplicatePriorityQueue(1024 * 2, 0.5);
					}

				};

			default :
				// Fallback
				throw new UnsupportedOperationException("Not supported " + queueingModel);
		}
	}


}