/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.controlflow.override;

import gnu.trove.TIntCollection;
import gnu.trove.iterator.TIntIterator;
import gnu.trove.list.TIntList;

import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import nl.tue.astar.AStarException;
import nl.tue.astar.AStarObserver;
import nl.tue.astar.Head;
import nl.tue.astar.Record;
import nl.tue.astar.Tail;
import nl.tue.astar.Trace;
import nl.tue.astar.impl.AbstractAStarThread;
import nl.tue.astar.impl.State;
import nl.tue.astar.impl.memefficient.StorageAwareDelegate;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.astar.petrinet.impl.PRecord;
import org.processmining.plugins.balancedconformance.cache.DataViewBuilder;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration.UnassignedMode;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResultImpl;
import org.processmining.plugins.balancedconformance.controlflow.adapter.AlignmentAdapter.PriorityQueueFactory;
import org.processmining.plugins.balancedconformance.controlflow.override.storage.ConsideredStatesStore;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapter;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapterCacheImpl;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapterChainingImpl;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapterSimpleCheckImpl;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapterWithoutPartialImpl;
import org.processmining.plugins.balancedconformance.dataflow.PartialDataAlignedTrace;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataState;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateHelper;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;
import org.processmining.plugins.balancedconformance.mapping.Variable;

/**
 * Returns the optimal control flow and data alignment by adding the data cost
 * of the current alignment prefix to the total cost of each record. This
 * implementation is NOT synchronized it should only be accesses by a single
 * thread at all times. It returns a {@link DRecord} instead of the normal
 * {@link PRecord} as we need to store some additional information.
 * 
 * @author fmannhardt
 * 
 * @param <H>
 * @param <T>
 */
public class DataAwareThread<H extends Head, T extends Tail> extends AbstractAStarThread<H, T> {

	private final class ConsideredStatesStoreNoOpImpl implements ConsideredStatesStore {
		public boolean contains(DRecord record) {
			return false;
		}

		public void add(DRecord record) {
		}
	}

	/**
	 * Stateful factory for {@link ControlFlowAlignmentResult} instances from a
	 * {@link DRecord}
	 */
	public static class PartialResultFactory {

		private final AbstractPDelegate<? extends Tail> delegate;
		private final Map<XEvent, XEventClass> mapEv2EventClass;
		private final Map<XEventClass, Integer> mapEvClass2Cost;
		private final XTrace xTrace;
		private final List<XEvent> mappedTrace;
		private final Trace trace;

		private DRecord[] history;

		public PartialResultFactory(AbstractPDelegate<? extends Tail> delegate,
				Map<XEvent, XEventClass> mapEv2EventClass, Map<XEventClass, Integer> mapEvClass2Cost, XTrace xTrace,
				List<XEvent> mappedTrace, Trace trace) {
			super();
			this.delegate = delegate;
			this.mapEv2EventClass = mapEv2EventClass;
			this.mapEvClass2Cost = mapEvClass2Cost;
			this.xTrace = xTrace;
			this.mappedTrace = mappedTrace;
			this.trace = trace;
		}

		public DRecord[] getHistory(DRecord r) {
			if (history == null) {
				history = r.getHistory();
			}
			return history;
		}

		public void newRecord() {
			history = null;
		}

		public ControlFlowAlignmentResult buildPartialFromCurrentRecord(DRecord r) {
			return new ControlFlowAlignmentResultImpl(r, delegate, mapEv2EventClass, mapEvClass2Cost, xTrace,
					mappedTrace, trace, true);
		}

	}

	protected final TraceMapping traceMapping;
	protected final LogMapping logMapping;

	protected final Map<XEvent, XEventClass> mapEv2EventClass;
	protected final Map<XEventClass, Integer> mapEvClass2Cost;

	protected final DataAlignmentAdapterChainingImpl dataAdapterChain;
	protected final PartialResultFactory partialResultFactory;
	private DataAlignmentAdapterCacheImpl cacheAdapter;

	protected int partialDataAlignmentCount;
	protected final DataStateStore dataStateStore;

	protected final H initialHead;
	protected ConsideredStatesStore consideredStatesStore;

	public DataAwareThread(StorageAwareDelegate<H, T> delegate, H initialHead, StorageHandler<H, T> storageHandler,
			Trace trace, int maxStates, DataAlignmentAdapter adapter, LogMapping logMapping, TraceMapping traceMapping,
			Map<XEvent, XEventClass> mapEv2EventClass, Map<XEventClass, Integer> mapEvClass2Cost,
			DataViewBuilder dataViewBuilder, boolean activateCache, boolean isUseImprovedMethod,
			ASynchronousMoveSorting sorting, PriorityQueueFactory queueFactory, DataStateStore dataStateStore) {
		super(delegate, trace, maxStates, storageHandler);
		this.initialHead = initialHead;
		this.logMapping = logMapping;
		this.traceMapping = traceMapping;
		this.mapEvClass2Cost = mapEvClass2Cost;
		this.mapEv2EventClass = mapEv2EventClass;
		this.partialResultFactory = traceMapping.getPartialResultFactory();

		this.sorting = sorting;
		this.dataStateStore = dataStateStore;

		this.dataAdapterChain = new DataAlignmentAdapterChainingImpl(adapter.getDataStateFactory());
		//TODO this is a ugly hack, move this code outside of the DataAwareThread
		if (!(adapter instanceof DataAlignmentAdapterWithoutPartialImpl)) {

			// MUST be first in chain
			if (activateCache && !logMapping.hasVirtualVariables()) {
				cacheAdapter = new DataAlignmentAdapterCacheImpl(dataViewBuilder, adapter.getDataStateFactory(),
						getAbstractPDelegate(), logMapping);
				dataAdapterChain.addAdapter(cacheAdapter);
				dataAdapterChain.addObserver(cacheAdapter);
			}

			if (isUseImprovedMethod) {
				dataAdapterChain.addAdapter(new DataAlignmentAdapterSimpleCheckImpl(logMapping, traceMapping, adapter
						.getDataStateFactory(), getAbstractPDelegate().getDelta(), getAbstractPDelegate(),
						dataStateStore));
			}

		}
		dataAdapterChain.addAdapter(adapter);

		/*
		 * Already considered states may improve, if a different path yields a
		 * better data alignment thus we do not maintain a closed list in the
		 * naive implementation.
		 */
		consideredStatesStore = new ConsideredStatesStoreNoOpImpl();

		setQueue(queueFactory.newInstance(dataViewBuilder, getAbstractPDelegate()));
		try {
			initializeQueue(initialHead);
		} catch (AStarException e) {
			throw new RuntimeException("Failed to initialize AStar Queue, something seriously went wrong!", e);
		}
	}

	public final int getCacheHit() {
		return cacheAdapter != null ? cacheAdapter.getCacheHits() : 0;
	}

	public final int getPartialDataAlignments() {
		return dataAdapterChain.getPartialDataAlignments();
	}

	public final int getCacheSize() {
		return cacheAdapter != null ? cacheAdapter.getCacheSize() : 0;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * nl.tue.astar.impl.AbstractAStarThread#initializeQueue(nl.tue.astar.Head)
	 */
	@Override
	protected void initializeQueue(H head) throws AStarException {
		// First, find the location of head
		final long index = storageHandler.getIndexOf(head);
		// note that the contains method may give false negatives. However,
		// it is assumed to be more expensive to synchronize on (algorithm) than
		// to
		// just recompute the tail.

		// create a record for this new head
		DRecordImpl newRec = new DRecordImpl(0, 0, null);
		newRec.setEstimatedRemainingCost(0, true);

		if (index < 0) {
			// the statespace doesn't contain a corresponding state
			final T tail = delegate.createInitialTail(head);
			final State<H, T> newState = new State<>(head, tail);

			storageHandler.storeStateForRecord(newState, newRec);

		} else {
			newRec.setState(index);
		}

		if (logMapping.getUnassignedMode() == UnassignedMode.DEFAULT) {
			DataState dataStateWithDefaults = dataAdapterChain.getDataStateFactory().newDataState();
			for (Entry<String, Variable> entry : logMapping.getVariables().entrySet()) {
				String variableName = entry.getKey();
				Variable variable = entry.getValue();
				Class<?> variableType = variable.getType();
				if (variable.getDefaultValue() != null) {
					DataStateHelper.storeValue(dataStateWithDefaults, variableName, variable.getDefaultValue(),
							variableType, logMapping.getStringDiscretizer());
				}
			}
			int dataStateIndex = dataStateStore.putIfAbsent(dataStateWithDefaults);
			newRec.setDataState(dataStateIndex);
		} else {
			int dataStateIndex = dataStateStore.putIfAbsent(DataStateStore.EMPTY_IMMUTABLE_STATE);
			newRec.setDataState(dataStateIndex);
		}

		if (queue.add(newRec)) {
			queuedStateCount++;
		}

	}

	@Override
	public void setQueueingModel(QueueingModel model) {
		// We take the priority queue supplied in the constructor
	}

	@SuppressWarnings("unchecked")
	protected final AbstractPDelegate<? extends Tail> getAbstractPDelegate() {
		return ((AbstractPDelegate<? extends Tail>) delegate);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * nl.tue.astar.impl.AbstractAStarThread#setConsidered(nl.tue.astar.Record)
	 */
	@Override
	protected final void setConsidered(Record record) {
		if (record != null) {
			getConsideredStateStore().add((DRecord) record);
			if (!observers.isEmpty()) {
				for (AStarObserver observer : observers) {
					observer.nodeVisited(record);
				}
			}
		}
	}

	protected final DRecord augmentRecord(DRecord newRecord) throws AStarException {

		// Second, augment with variable assignments
		try {
			PartialDataAlignedTrace partialDataAlignmentState = dataAdapterChain.retrievePartialDataAlignment(
					traceMapping, partialResultFactory, newRecord);

			if (partialDataAlignmentState == PartialDataAlignedTrace.IMPOSSIBLE) {
				// This move is not allowed as there is no data alignment for the resulting prefix
				return null;
			} else {

				int dataIndex;
				if (partialDataAlignmentState.getVariableAssignment() != null) {
					// Add data state to storage
					dataIndex = dataStateStore.putIfAbsent(partialDataAlignmentState.getVariableAssignment());
				} else {
					// Use existing index
					dataIndex = partialDataAlignmentState.getDataStateIndex();
				}

				newRecord.setDataCost(partialDataAlignmentState.getDataCost());
				newRecord.setDataState(dataIndex);

				return newRecord;
			}

		} catch (DataAlignmentException e1) {
			throw new AStarException(e1);
		}
	}

	protected final ConsideredStatesStore getConsideredStateStore() {
		return consideredStatesStore;
	}

	@Override
	protected void processMovesForRecordWithUpToDateTail(Record rec, H head, T tail, int stopAt, double timeLimit,
			long endTime) throws AStarException {

		setConsidered(rec);

		// move model only
		TIntList enabled = head.getModelMoves(rec, delegate);

		TIntCollection nextEvents = rec.getNextEvents(delegate, trace);
		TIntIterator evtIt = nextEvents.iterator();
		//		int nextEvent = ((DRecord) rec).getNextEvent(delegate, trace);

		int activity = NOMOVE;

		while (evtIt.hasNext()) {
			int nextEvent = evtIt.next();
			//		if (nextEvent < trace.getSize()) {
			TIntList ml = null;

			// move both log and model synchronously;
			activity = trace.get(nextEvent);
			ml = head.getSynchronousMoves(rec, delegate, enabled, activity);
			TIntIterator it = ml.iterator();
			while (it.hasNext()) {
				processMove(head, tail, rec, it.next(), nextEvent, activity);
			}

			// sorting == ASynchronousMoveSorting.LOGMOVEFIRST implies
			// logMove only after initial move, sync move or log move.
			if (isValidMoveOnLog(rec, nextEvent, activity, enabled, ml)) {
				// allow move on log only if the previous move was
				// 1) the initial move (rec.getPredecessor() == null
				// 2) a synchronous move
				// 3) a log move.
				processMove(head, tail, rec, NOMOVE, nextEvent, activity);
			}
		}

		// sorting == ASynchronousMoveSorting.MODELMOVEFIRST implies
		// modelMove only after initial move, sync move or model move.
		if (isValidMoveOnModel(rec, null, activity, enabled)) {
			// allow move on model only if previous move was:
			// 1) the initial move (rec.getPredecessor() == null
			// 2) a synchronous move
			// 3) a move on model.
			TIntIterator it = enabled.iterator();
			while (it.hasNext()) {
				// move model
				processMove(head, tail, rec, it.next(), NOMOVE, NOMOVE);
			}
		}

	}

	@Override
	protected void processMove(H head, T tail, Record rec, int modelMove, int movedEvent, int activity)
			throws AStarException {

		// First, construct the next head from the old head
		final H newHead = computeNextHead(rec, head, modelMove, movedEvent, activity);
		final long index;
		try {
			// Check whether the new head is already in search space
			index = storageHandler.getIndexOf(newHead);
		} catch (Exception e) {
			throw new AStarException(e);
		}

		// First, create control-flow record
		DRecord newRecord = ((DRecord) rec).getNextRecord(delegate, trace, newHead, index, modelMove, movedEvent,
				activity);
		DRecord newRec = augmentRecord(newRecord);
		// Skip all hopeless states
		if (newRec == null) {
			return;
		}

		traversedArcCount++;

		final T newTail;
		if (index >= 0) {
			// so far, we did the cheap stuff, now get the new tail from
			// storage, knowing that
			// a state exists at index
			int h;
			try {
				h = storageHandler.getEstimate(newHead, index);
			} catch (Exception e) {
				throw new AStarException(e);
			}

			newRec.setState(index);
			newRec.setEstimatedRemainingCost(h, true);
			enqueueRecord(newRec);

			// edge from oldRec to newRec traversed.
			if (!observers.isEmpty()) {
				for (AStarObserver observer : observers) {
					observer.edgeTraversed(rec, newRec);
				}
			}
			return;
		}

		// the statespace doesn't contain a corresponding state, hence we need
		// to compute the tail.
		newTail = computeNewTail(rec, tail, newHead, modelMove, movedEvent, activity);

		if (!newTail.canComplete()) {
			return;
		}

		int h = newTail.getEstimatedCosts(delegate, newHead);

		// Check if the head is in the store and add if it isn't.
		final State<H, T> newState = new State<>(newHead, newTail);

		try {
			storageHandler.storeStateForRecord(newState, newRec);
		} catch (Exception e) {
			throw new AStarException(e);
		}

		newRec.setEstimatedRemainingCost(h, true);
		enqueueRecord(newRec);

		// edge from oldRec to newRec traversed.
		if (!observers.isEmpty()) {
			for (AStarObserver observer : observers) {
				observer.edgeTraversed(rec, newRec);
			}
		}

	}

	protected final void enqueueRecord(DRecord newRec) {
		if (queue.add(newRec)) {
			queuedStateCount++;
		}
	}

	public final long getDataStateCount() {
		return dataStateStore.size();
	}

	public final boolean isQueueEmpty() {
		return queue.isEmpty();
	}

}