/*
 * Mainly copied from nl.tue.astar.util.AbstractFastLookupPriorityQueue
 */
package org.processmining.plugins.balancedconformance.controlflow.override.queues.duplicates;

import java.util.Arrays;
import java.util.Queue;

import nl.tue.astar.Head;
import nl.tue.astar.Record;
import nl.tue.astar.util.AbstractFastLookupPriorityQueue;
import nl.tue.astar.util.FastLookupPriorityQueue;

/**
 * Implementation of a {@link FastLookupPriorityQueue} that allows adding
 * multiple {@link Record} objects that point to the same state (i.e.
 * {@link Head}). It does not support the methods {@link #contains(Record)}
 * though. Most code taken from {@link AbstractFastLookupPriorityQueue}, but it
 * seemed better to not subclass it, as we don't need some fields.
 * 
 * @author F. Mannhardt
 * 
 */
public abstract class DuplicateStatesFastLookupPriorityQueue implements FastLookupPriorityQueue {

	/**
	 * Priority queue represented as a balanced binary heap: the two children of
	 * queue[n] are queue[2*n+1] and queue[2*(n+1)]. The priority queue is
	 * ordered by the record's natural ordering: For each node n in the heap and
	 * each descendant d of n, n <= d. The element with the best value is in
	 * queue[0], assuming the queue is nonempty.
	 */
	protected Record[] queue;

	/**
	 * The number of elements in the priority queue.
	 */
	protected int size = 0;

	private int maxCost;

	public DuplicateStatesFastLookupPriorityQueue(int initialCapacity) {
		this.queue = new Record[initialCapacity];
	}

	public void setMaxCost(int maxCost) {
		this.maxCost = maxCost;
	}

	public int getMaxCost() {
		return maxCost;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return Arrays.toString(queue);
	}

	abstract protected boolean isBetter(Record r1, Record r2);

	/*
	 * (non-Javadoc)
	 * 
	 * @see nl.tue.astar.util.FastLookupPriorityQueue#add(nl.tue.astar.Record)
	 */
	public boolean add(Record newE) {
		offer(newE);
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see nl.tue.astar.util.FastLookupPriorityQueue#isEmpty()
	 */
	public boolean isEmpty() {
		return size() == 0;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * nl.tue.astar.util.FastLookupPriorityQueue#contains(nl.tue.astar.Record)
	 */
	public Record contains(Record newRec) {
		throw new UnsupportedOperationException("This queue may contain duplicates, so contains is not defined!");
	}

	/**
	 * Increases the capacity of the array.
	 * 
	 * @param minCapacity
	 *            the desired minimum capacity
	 */
	private void grow(int minCapacity) {
		if (minCapacity < 0) // overflow
			throw new OutOfMemoryError();
		int oldCapacity = queue.length;
		// Double size if small; else grow by 50%
		int newCapacity = ((oldCapacity < 64) ? ((oldCapacity + 1) * 2) : ((oldCapacity / 4) * 5));
		if (newCapacity < 0) // overflow
			newCapacity = Integer.MAX_VALUE;
		if (newCapacity < minCapacity)
			newCapacity = minCapacity;
		queue = Arrays.copyOf(queue, newCapacity);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see nl.tue.astar.util.FastLookupPriorityQueue#peek()
	 */
	public Record peek() {
		if (size == 0)
			return null;
		return queue[0];
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see nl.tue.astar.util.FastLookupPriorityQueue#size()
	 */
	public int size() {
		return size;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see nl.tue.astar.util.FastLookupPriorityQueue#poll()
	 */
	public Record poll() {
		if (size == 0)
			return null;
		int s = --size;
		Record result = queue[0];
		Record x = queue[s];
		queue[s] = null;

		if (s != 0)
			siftDown(0, x);
		return result;
	}

	/**
	 * Inserts the specified element into this priority queue.
	 * 
	 * @return {@code true} (as specified by {@link Queue#offer})
	 * @throws ClassCastException
	 *             if the specified element cannot be compared with elements
	 *             currently in this priority queue according to the priority
	 *             queue's ordering
	 * @throws NullPointerException
	 *             if the specified element is null
	 */
	private void offer(Record e) {
		if (e == null)
			throw new NullPointerException();
		int i = size;
		if (i >= queue.length)
			grow(i + 1);
		size = i + 1;
		if (i == 0) {
			queue[0] = e;
		} else
			siftUp(i, e);
	}

	/**
	 * Inserts item x at position k, maintaining heap invariant by promoting x
	 * up the tree until it is greater than or equal to its parent, or is the
	 * root.
	 * 
	 * @param k
	 * @param x
	 *            the item to insert
	 */
	private void siftUp(int k, Record x) {
		while (k > 0) {
			int parent = (k - 1) >>> 1;
			Record e = queue[parent];
			if (!isBetter(x, e)) {
				break;
			}
			queue[k] = e;
			k = parent;
		}
		queue[k] = x;
	}

	/**
	 * Inserts item x at position k, maintaining heap invariant by demoting x
	 * down the tree repeatedly until it is less than or equal to its children
	 * or is a leaf.
	 * 
	 * @param k
	 *            the position to fill
	 * @param x
	 *            the item to insert
	 */
	private void siftDown(int k, Record x) {
		int half = size >>> 1;
		while (k < half) {
			int child = (k << 1) + 1;
			Record c = queue[child];
			int right = child + 1;
			if (right < size && isBetter(queue[right], c))
				c = queue[child = right];

			if (!isBetter(c, x))
				break;

			queue[k] = c;
			k = child;
		}
		queue[k] = x;
	}

}
