/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.dataflow;

import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;
import org.processmining.plugins.balancedconformance.controlflow.override.DataAwareThread.PartialResultFactory;
import org.processmining.plugins.balancedconformance.dataflow.MILPProblemBuilder.FullDataAlignedTrace;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateFactory;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.dataflow.exception.InfeasibleMILPProblem;
import org.processmining.plugins.balancedconformance.dataflow.exception.MILPProblemBuildException;
import org.processmining.plugins.balancedconformance.dataflow.exception.UnreliableDataAlignmentException;
import org.processmining.plugins.balancedconformance.dataflow.exception.UnreliableDataAlignmentException.UnreliableReason;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;
import org.processmining.plugins.balancedconformance.result.DataAlignedTrace;
import org.processmining.plugins.balancedconformance.result.DataAlignedTraceImpl;

import net.sf.javailp.SolverFactory;

/**
 * {@link DataAlignmentAdapter} using a MILP solver
 * 
 * @author F. Mannhardt
 * 
 */
public class DataAlignmentAdapterMILPImpl extends AbstractDataAlignmentAdapter implements ExpensiveAdapter {

	protected final MILPProblemBuilder singleTraceDataAlignment;
	private final SolverFactory ilpSolver;

	public DataAlignmentAdapterMILPImpl(LogMapping mapping, SolverFactory ilpSolver, DataStateFactory dataStateFactory,
			double aStarDelta, boolean includeVirtualVariablesInTrace, boolean usePartialMILPProblems) {
		super(mapping, dataStateFactory);
		this.ilpSolver = ilpSolver;
		this.singleTraceDataAlignment = new MILPProblemBuilderImpl(mapping, ilpSolver, dataStateFactory, aStarDelta,
				includeVirtualVariablesInTrace, usePartialMILPProblems);
	}

	@Override
	public DataAlignedTrace retrieveFullDataAlignment(TraceMapping traceMapping,
			ControlFlowAlignmentResult controlFlowAlignment) throws DataAlignmentException {
		try {
			FullDataAlignedTrace fullDataAlignment = singleTraceDataAlignment.createAlignment(traceMapping,
					controlFlowAlignment);
			if (fullDataAlignment == FullDataAlignedTrace.IMPOSSIBLE) {
				throw new UnreliableDataAlignmentException(
						"Data alignment impossible! It was not possible to validate the variable assignment for the control-flow found."
								+ " Maybe there is some numerical instability in the ILP solving that causes incorrect results."
								+ " Check whether appropriate bounds for the variables have been set.\n\n"
								+ controlFlowAlignment.toString(), null, UnreliableReason.IMPOSSIBLE, traceMapping.getTrace());
			} else {
				return new DataAlignedTraceImpl(fullDataAlignment.getLogTracePrefix(),
						fullDataAlignment.getProcessTracePrefix(), controlFlowAlignment.getControlFlowCost(),
						fullDataAlignment.getDataCost());
			}
		} catch (InfeasibleMILPProblem e) {
			throw new UnreliableDataAlignmentException(
					"Data alignment impossible! It was not possible to validate the variable assignment for the control-flow found."
							+ " Maybe there is some numerical instability in the ILP solving that causes incorrect results."
							+ " Check whether appropriate bounds for the variables have been set.\n\n Trying to solve the following ILP problem:\n"
							+ e.getMessage(), null, UnreliableReason.IMPOSSIBLE, traceMapping.getTrace());
		} catch (MILPProblemBuildException e) {
			throw new DataAlignmentException("Failed to build data alignment.", e);
		}
	}

	@Override
	public PartialDataAlignedTrace retrievePartialDataAlignment(TraceMapping traceMapping,
			PartialResultFactory partialResultBuilder, DRecord record) throws DataAlignmentException {
		try {
			return singleTraceDataAlignment.createPartialAlignment(traceMapping,
					partialResultBuilder.buildPartialFromCurrentRecord(record));
		} catch (MILPProblemBuildException e) {
			throw new DataAlignmentException("Failed to build data alignment.", e);
		}
	}

	public void close() {
		if (ilpSolver instanceof AutoCloseable) {
			try {
				((AutoCloseable) ilpSolver).close();
			} catch (Exception e) {
				throw new RuntimeException(e);
			}	
		}
	}

}