package org.processmining.plugins.balancedconformance.dataflow;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.datapetrinets.exception.NumericValueRequired;
import org.processmining.datapetrinets.expression.Evaluator;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.expression.Printer;
import org.processmining.datapetrinets.expression.syntax.ExprAnd;
import org.processmining.datapetrinets.expression.syntax.ExprAtLeast;
import org.processmining.datapetrinets.expression.syntax.ExprAtMost;
import org.processmining.datapetrinets.expression.syntax.ExprDiv;
import org.processmining.datapetrinets.expression.syntax.ExprEqual;
import org.processmining.datapetrinets.expression.syntax.ExprFunction;
import org.processmining.datapetrinets.expression.syntax.ExprGreaterThan;
import org.processmining.datapetrinets.expression.syntax.ExprLessThan;
import org.processmining.datapetrinets.expression.syntax.ExprLitBoolean;
import org.processmining.datapetrinets.expression.syntax.ExprLitDouble;
import org.processmining.datapetrinets.expression.syntax.ExprLitInteger;
import org.processmining.datapetrinets.expression.syntax.ExprLitNull;
import org.processmining.datapetrinets.expression.syntax.ExprLitString;
import org.processmining.datapetrinets.expression.syntax.ExprMinus;
import org.processmining.datapetrinets.expression.syntax.ExprMult;
import org.processmining.datapetrinets.expression.syntax.ExprNegation;
import org.processmining.datapetrinets.expression.syntax.ExprNot;
import org.processmining.datapetrinets.expression.syntax.ExprNotEqual;
import org.processmining.datapetrinets.expression.syntax.ExprOr;
import org.processmining.datapetrinets.expression.syntax.ExprPlus;
import org.processmining.datapetrinets.expression.syntax.ExprRoot;
import org.processmining.datapetrinets.expression.syntax.ExprVariable;
import org.processmining.datapetrinets.expression.syntax.ExpressionParserDefaultVisitor;
import org.processmining.datapetrinets.expression.syntax.ExpressionParserTreeConstants;
import org.processmining.datapetrinets.expression.syntax.ExpressionVisitorException;
import org.processmining.datapetrinets.expression.syntax.Node;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration.UnassignedMode;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateHelper;
import org.processmining.plugins.balancedconformance.dataflow.lp.ILPProblem;
import org.processmining.plugins.balancedconformance.mapping.NotLinearException;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizer;
import org.processmining.plugins.balancedconformance.mapping.Variable;

import com.google.common.collect.ImmutableList;

import gnu.trove.map.TObjectIntMap;
import net.sf.javailp.Constraint;
import net.sf.javailp.Linear;
import net.sf.javailp.Operator;
import net.sf.javailp.Term;

/**
 * Class that encapsulates the the conversion between a {@link GuardExpression}
 * and an ILP problem.
 * 
 * @author F. Mannhardt
 * 
 */
public final class IlpExpressionTemplate {

	private static final class VariableCoefficient {

		String variable;
		double coefficient;

		public static VariableCoefficient create(String variable, double coefficient) {
			VariableCoefficient variableCoefficient = new VariableCoefficient();
			variableCoefficient.variable = variable;
			variableCoefficient.coefficient = coefficient;
			return variableCoefficient;
		}
	}

	private static final class LinearForm {

		final List<VariableCoefficient> coefficients = new ArrayList<IlpExpressionTemplate.VariableCoefficient>();

	}

	private static final class Info {

		boolean inNegation;
		List<String> orVariables;
		LinearForm linearForm;

		boolean isInOR() {
			return orVariables != null && !orVariables.isEmpty();
		}

		static Info create(boolean inNegation) {
			Info data = new Info();
			data.inNegation = inNegation;
			return data;
		}

		static Info create(boolean inNegation, List<String> orVariables) {
			Info data = Info.create(inNegation);
			data.orVariables = orVariables;
			return data;
		}

		static Info create(boolean inNegation, List<String> orVariables, LinearForm linearForm) {
			Info data = Info.create(inNegation, orVariables);
			data.linearForm = linearForm;
			return data;
		}

		static Info create() {
			return new Info();
		}

	}

	private final class ExpressionConversionVisitor extends Evaluator.Visitor {

		private Info info(Object data) {
			if (data == null) {
				return Info.create();
			} else {
				return (Info) data;
			}
		}

		private Info clone(Info data) {
			return Info.create(data.inNegation, data.orVariables, data.linearForm);
		}

		@Override
		public Object visit(ExprRoot node, Object data) throws ExpressionVisitorException {
			if (node.jjtGetNumChildren() > 1) {
				throw new ExpressionVisitorException("Invalid expression " + Printer.printCanonical(node)
						+ " should not have been parsed! Top level element is only allowed to have one child.");
			}
			return node.jjtGetChild(0).jjtAccept(this, data);
		}

		@Override
		public Object visit(ExprOr node, Object data) throws ExpressionVisitorException {
			Info info = info(data);
			if (info.inNegation) {
				visitAnd(node, info);
			} else {
				visitOr(node, info);
			}
			return null;
		}

		@Override
		public Object visit(ExprAnd node, Object data) throws ExpressionVisitorException {
			Info info = info(data);
			if (info.inNegation) {
				visitOr(node, info);
			} else {
				visitAnd(node, info);
			}
			return null;
		}

		private void visitOr(Node node, Info data) throws ExpressionVisitorException {

			boolean hasDirectOrChildren = hasDirectOrChildren(node, data);

			if (hasDirectOrChildren) {

				List<String> orVariables = new ArrayList<>();
				List<Node> orNodes = new ArrayList<>();
				orNodes.add(node);
				// Collapse multiple OR constraints into a single constraint
				collectDirectlyNestedORs(node, data, orNodes);

				for (Node orNode : orNodes) {

					for (int i = 0; i < orNode.jjtGetNumChildren(); i++) {
						Node child = orNode.jjtGetChild(i);
						if (!isOrNode(child, data)) {
							String variableForOr = createORVariable();
							List<String> variablesForChild = createVariableList(data, variableForOr);
							orVariables.add(variableForOr);
							child.jjtAccept(this, Info.create(data.inNegation, variablesForChild));
						}
					}

				}

				addConstraintForOR(orVariables);

			} else {
				// Add two OR variables and link them through a constraint
				String variableForLhs = createORVariable();
				List<String> lhsOrVariables = createVariableList(data, variableForLhs);
				node.jjtGetChild(0).jjtAccept(this, Info.create(data.inNegation, lhsOrVariables));
				String variableForRhs = createORVariable();
				List<String> rhsOrVariables = createVariableList(data, variableForRhs);
				node.jjtGetChild(1).jjtAccept(this, Info.create(data.inNegation, rhsOrVariables));
				addConstraintForOR(ImmutableList.of(variableForLhs, variableForRhs));
			}
		}

		private void collectDirectlyNestedORs(Node node, Info data, List<Node> directlyNestedOrNodes)
				throws ExpressionVisitorException {

			Node lhs = node.jjtGetChild(0);
			Node rhs = node.jjtGetChild(1);

			if (isOrNode(lhs, data) && isOrNode(rhs, data)) {
				directlyNestedOrNodes.add(lhs);
				collectDirectlyNestedORs(lhs, data, directlyNestedOrNodes);
				directlyNestedOrNodes.add(rhs);
				collectDirectlyNestedORs(rhs, data, directlyNestedOrNodes);
			} else if (isOrNode(lhs, data)) {
				directlyNestedOrNodes.add(lhs);
				collectDirectlyNestedORs(lhs, data, directlyNestedOrNodes);
			} else if (isOrNode(rhs, data)) {
				directlyNestedOrNodes.add(rhs);
				collectDirectlyNestedORs(rhs, data, directlyNestedOrNodes);
			}

		}

		private boolean isOrNode(Node node, Info data) {
			return node instanceof ExprOr || (data.inNegation && node instanceof ExprAnd);
		}

		private boolean hasDirectOrChildren(Node node, Info data) {
			Node currentNode = node;
			Node lhs = currentNode.jjtGetChild(0);
			Node rhs = currentNode.jjtGetChild(1);
			return isOrNode(lhs, data) || isOrNode(rhs, data);
		}

		private final void addConstraintForOR(Collection<String> variablesOr) {
			Linear constraintForOR = new Linear();
			for (String orVar : variablesOr) {
				booleanVariables.add(orVar);
				constraintForOR.add(1, orVar);
			}
			extraOrConstraints
					.add(createConstraint(constraintForOR, Operator.LE, Double.valueOf(variablesOr.size() - 1)));
		}

		private List<String> createVariableList(Info data, String newVariable) {
			List<String> variableList;
			if (data.orVariables == null) {
				variableList = new ArrayList<>();
			} else {
				variableList = new ArrayList<>(data.orVariables);
			}
			variableList.add(newVariable);
			return variableList;
		}

		private void visitAnd(Node node, Info data) throws ExpressionVisitorException {
			node.jjtGetChild(0).jjtAccept(this, clone(data));
			node.jjtGetChild(1).jjtAccept(this, clone(data));
		}

		public Object visit(ExprNegation node, Object data) throws ExpressionVisitorException {
			Info info = info(data);
			Node child = node.jjtGetChild(0);
			if (child instanceof ExprVariable) {
				String variableName = (String) ((ExprVariable) child).jjtGetValue();
				info.linearForm.coefficients.add(VariableCoefficient.create(variableName, -1));
				return 0.0d;
			} else {
				return super.visit(node, Info.create(info.inNegation, info.orVariables, null));
			}
		}

		public Object visit(ExprDiv node, Object data) throws ExpressionVisitorException {
			Info info = info(data);
			Node lhsNode = node.jjtGetChild(0);
			Node rhsNode = node.jjtGetChild(1);
			if (lhsNode instanceof ExprVariable && rhsNode instanceof ExprVariable) {
				String lhsVariable = (String) ((ExprVariable) lhsNode).jjtGetValue();
				String rhsVariable = (String) ((ExprVariable) rhsNode).jjtGetValue();
				throw new NotLinearException(
						String.format("Non-linear constraint %s * %s is not supported!", lhsVariable, rhsVariable));
			} else if (lhsNode instanceof ExprVariable) {
				String variableName = (String) ((ExprVariable) lhsNode).jjtGetValue();
				Object rhsValue = rhsNode.jjtAccept(this, clone(info));
				if (rhsValue instanceof Number) {
					info.linearForm.coefficients
							.add(VariableCoefficient.create(variableName, 1.0d / ((Number) rhsValue).doubleValue()));
					return 0.0d;
				} else {
					throw new NumericValueRequired(
							String.format("Constraint %s - %s is not supported!", variableName, rhsValue));
				}
			} else if (rhsNode instanceof ExprVariable) {
				String rhsVariable = (String) ((ExprVariable) lhsNode).jjtGetValue();
				throw new NotLinearException(
						String.format("Non-linear constraint %s / %s is not supported!", lhsNode, rhsVariable));
			} else {
				return super.visit(node, data);
			}
		}

		public Object visit(ExprMult node, Object data) throws ExpressionVisitorException {
			Info info = info(data);
			Node lhsNode = node.jjtGetChild(0);
			Node rhsNode = node.jjtGetChild(1);
			if (lhsNode instanceof ExprVariable && rhsNode instanceof ExprVariable) {
				String lhsVariable = (String) ((ExprVariable) lhsNode).jjtGetValue();
				String rhsVariable = (String) ((ExprVariable) rhsNode).jjtGetValue();
				throw new NotLinearException(
						String.format("Non-linear constraint %s * %s is not supported!", lhsVariable, rhsVariable));
			} else if (lhsNode instanceof ExprVariable) {
				String variableName = (String) ((ExprVariable) lhsNode).jjtGetValue();
				Object rhsValue = rhsNode.jjtAccept(this, clone(info));
				if (rhsValue instanceof Number) {
					info.linearForm.coefficients
							.add(VariableCoefficient.create(variableName, ((Number) rhsValue).doubleValue()));
					return 0.0d;
				} else {
					throw new NumericValueRequired(
							String.format("Constraint %s - %s is not supported!", variableName, rhsValue));
				}
			} else if (rhsNode instanceof ExprVariable) {
				String variableName = (String) ((ExprVariable) rhsNode).jjtGetValue();
				Object lhsValue = lhsNode.jjtAccept(this, clone(info));
				if (lhsValue instanceof Number) {
					info.linearForm.coefficients
							.add(VariableCoefficient.create(variableName, ((Number) lhsValue).doubleValue()));
					return 0.0d;
				} else {
					throw new NumericValueRequired(
							String.format("Constraint %s - %s is not supported!", variableName, lhsValue));
				}
			} else {
				return super.visit(node, data);
			}
		}

		public Object visit(ExprMinus node, Object data) throws ExpressionVisitorException {
			Info info = info(data);
			Node lhsNode = node.jjtGetChild(0);
			Node rhsNode = node.jjtGetChild(1);
			if (lhsNode instanceof ExprVariable && rhsNode instanceof ExprVariable) {
				String lhsVariable = (String) ((ExprVariable) lhsNode).jjtGetValue();
				String rhsVariable = (String) ((ExprVariable) rhsNode).jjtGetValue();
				info.linearForm.coefficients.add(VariableCoefficient.create(lhsVariable, 1));
				info.linearForm.coefficients.add(VariableCoefficient.create(rhsVariable, -1));
				return 0.0d;
			} else if (lhsNode instanceof ExprVariable) {
				String variableName = (String) ((ExprVariable) lhsNode).jjtGetValue();
				info.linearForm.coefficients.add(VariableCoefficient.create(variableName, 1));
				Object rhsValue = rhsNode.jjtAccept(this, clone(info));
				if (rhsValue instanceof Number) {
					return -((Number) rhsValue).doubleValue();
				} else {
					throw new NumericValueRequired(
							String.format("Constraint %s - %s is not supported!", variableName, rhsValue));
				}
			} else if (rhsNode instanceof ExprVariable) {
				String variableName = (String) ((ExprVariable) rhsNode).jjtGetValue();
				info.linearForm.coefficients.add(VariableCoefficient.create(variableName, -1));
				return lhsNode.jjtAccept(this, clone(info));
			} else {
				return super.visit(node, data);
			}
		}

		public Object visit(ExprPlus node, Object data) throws ExpressionVisitorException {
			Info info = info(data);
			Node lhsNode = node.jjtGetChild(0);
			Node rhsNode = node.jjtGetChild(1);
			if (lhsNode instanceof ExprVariable && rhsNode instanceof ExprVariable) {
				String lhsVariable = (String) ((ExprVariable) lhsNode).jjtGetValue();
				String rhsVariable = (String) ((ExprVariable) rhsNode).jjtGetValue();
				info.linearForm.coefficients.add(VariableCoefficient.create(lhsVariable, 1));
				info.linearForm.coefficients.add(VariableCoefficient.create(rhsVariable, 1));
				return 0.0d;
			} else if (lhsNode instanceof ExprVariable) {
				String variableName = (String) ((ExprVariable) lhsNode).jjtGetValue();
				info.linearForm.coefficients.add(VariableCoefficient.create(variableName, 1));
				return rhsNode.jjtAccept(this, clone(info));
			} else if (rhsNode instanceof ExprVariable) {
				String variableName = (String) ((ExprVariable) rhsNode).jjtGetValue();
				info.linearForm.coefficients.add(VariableCoefficient.create(variableName, 1));
				return lhsNode.jjtAccept(this, clone(info));
			} else {
				return super.visit(node, data);
			}
		}

		public Object visit(ExprLitNull node, Object data) {
			return Double.valueOf(Double.NEGATIVE_INFINITY);
		}

		public Object visit(ExprLitBoolean node, Object data) {
			Object value = super.visit(node, data);
			return (Boolean) value ? 1.0d : 0.0d;
		}

		public Object visit(ExprLitString node, Object data) {
			Object value = super.visit(node, data);
			Object possibleDate = tryParseDateToMillis(value);
			if (possibleDate instanceof Long) {
				return ((Long) value).doubleValue();
			} else {
				return Double.valueOf(discretizer.convertToInt((String) value));
			}
		}

		public Object visit(ExprLitDouble node, Object data) {
			Object value = super.visit(node, data);
			return ((Double) value).doubleValue();
		}

		public Object visit(ExprLitInteger node, Object data) {
			Object value = super.visit(node, data);
			return ((Long) value).doubleValue();
		}

		@Override
		public Object visit(ExprNot node, Object data) throws ExpressionVisitorException {
			Info info = info(data);
			return node.jjtGetChild(0).jjtAccept(this,
					Info.create(!info.inNegation, info.orVariables, info.linearForm));
		}

		@Override
		public Object visit(ExprVariable node, Object data) throws ExpressionVisitorException {
			Info info = info(data);
			if (info.linearForm != null) {
				String variableName = (String) node.jjtGetValue();
				info.linearForm.coefficients.add(VariableCoefficient.create(variableName, 1));
				return 0.0d; // variable does not contribute to the constants
			}
			throw new NotLinearException("Expression involves a non-linear construct!");
		}

		public Object visit(ExprFunction node, Object data) {
			throw new UnsupportedOperationException("No support for functions yet!");
		}

		private Double visitBinaryComparison(Info info, Node lhsNode, Node rhsNode, Linear constraint,
				LinearForm leftNormalForm, LinearForm rightNormalForm) throws ExpressionVisitorException {
			Object lhsConstant = lhsNode.jjtAccept(this,
					Info.create(info.inNegation, info.orVariables, leftNormalForm));
			Object rhsConstant = rhsNode.jjtAccept(this,
					Info.create(info.inNegation, info.orVariables, rightNormalForm));

			for (VariableCoefficient variableCoefficient : leftNormalForm.coefficients) {
				constraint.add(variableCoefficient.coefficient, variableCoefficient.variable);
			}
			for (VariableCoefficient variableCoefficient : rightNormalForm.coefficients) {
				constraint.add(-variableCoefficient.coefficient, variableCoefficient.variable);
			}

			if (rhsConstant == null || lhsConstant == null) {
				return null;
			} else {
				return ((Double) rhsConstant) - ((Double) lhsConstant);
			}
		}

		@Override
		public Object visit(ExprEqual node, Object data) throws ExpressionVisitorException {
			Info info = info(data);

			Node lhsNode = node.jjtGetChild(0);
			Node rhsNode = node.jjtGetChild(1);

			Linear constraint = new Linear();
			LinearForm leftNormalForm = new LinearForm();
			LinearForm rightNormalForm = new LinearForm();

			Double rhsConstant = visitBinaryComparison(info, lhsNode, rhsNode, constraint, leftNormalForm,
					rightNormalForm);

			if (hasVariables(leftNormalForm)) {
				if (info.inNegation) {
					addNotEqualsConstraint(info, constraint, rhsConstant);
				} else {
					addEqualsConstraint(info, constraint, rhsConstant);
				}
			}

			return null;
		}

		private boolean hasVariables(LinearForm normalform) {
			return !normalform.coefficients.isEmpty();
		}

		@Override
		public Object visit(ExprNotEqual node, Object data) throws ExpressionVisitorException {

			Info info = info(data);

			Node lhsNode = node.jjtGetChild(0);
			Node rhsNode = node.jjtGetChild(1);

			Linear constraint = new Linear();
			LinearForm leftNormalForm = new LinearForm();
			LinearForm rightNormalForm = new LinearForm();

			Double rhsConstant = visitBinaryComparison(info, lhsNode, rhsNode, constraint, leftNormalForm,
					rightNormalForm);

			if (hasVariables(leftNormalForm)) {
				if (info.inNegation) {
					addEqualsConstraint(info, constraint, rhsConstant);
				} else {
					addNotEqualsConstraint(info, constraint, rhsConstant);
				}
			}

			return null;
		}

		@Override
		public Object visit(ExprLessThan node, Object data) throws ExpressionVisitorException {

			Info info = info(data);

			Node lhsNode = node.jjtGetChild(0);
			Node rhsNode = node.jjtGetChild(1);

			Linear constraint = new Linear();
			LinearForm leftNormalForm = new LinearForm();
			LinearForm rightNormalForm = new LinearForm();

			Double rhsConstant = visitBinaryComparison(info, lhsNode, rhsNode, constraint, leftNormalForm,
					rightNormalForm);

			if (rhsConstant == null) {
				throw new ExpressionVisitorException("Cannot use NULL in numeric comparison!");
			} else {
				if (hasVariables(leftNormalForm)) {
					if (info.inNegation) {
						addGreaterThanConstraint(info, constraint, rhsConstant);
					} else {
						addLessThanConstraint(info, constraint, rhsConstant);
					}
				}
			}

			return null;
		}

		@Override
		public Object visit(ExprAtMost node, Object data) throws ExpressionVisitorException {

			Info info = info(data);

			Node lhsNode = node.jjtGetChild(0);
			Node rhsNode = node.jjtGetChild(1);

			Linear constraint = new Linear();
			LinearForm leftNormalForm = new LinearForm();
			LinearForm rightNormalForm = new LinearForm();

			Double rhsConstant = visitBinaryComparison(info, lhsNode, rhsNode, constraint, leftNormalForm,
					rightNormalForm);

			if (rhsConstant == null) {
				throw new ExpressionVisitorException("Cannot use NULL in numeric comparison!");
			} else {
				if (hasVariables(leftNormalForm)) {
					if (info.inNegation) {
						addAtLeastConstraint(info, constraint, rhsConstant);
					} else {
						addAtMostConstraint(info, constraint, rhsConstant);
					}
				}
			}

			return null;
		}

		@Override
		public Object visit(ExprGreaterThan node, Object data) throws ExpressionVisitorException {

			Info info = info(data);

			Node lhsNode = node.jjtGetChild(0);
			Node rhsNode = node.jjtGetChild(1);

			Linear constraint = new Linear();
			LinearForm leftNormalForm = new LinearForm();
			LinearForm rightNormalForm = new LinearForm();

			Double rhsConstant = visitBinaryComparison(info, lhsNode, rhsNode, constraint, leftNormalForm,
					rightNormalForm);

			if (rhsConstant == null) {
				throw new ExpressionVisitorException("Cannot use NULL in numeric comparison!");
			} else {
				if (hasVariables(leftNormalForm)) {
					if (info.inNegation) {
						addLessThanConstraint(info, constraint, rhsConstant);
					} else {
						addGreaterThanConstraint(info, constraint, rhsConstant);
					}
				}
			}

			return null;
		}

		@Override
		public Object visit(ExprAtLeast node, Object data) throws ExpressionVisitorException {
			Info info = info(data);

			Node lhsNode = node.jjtGetChild(0);
			Node rhsNode = node.jjtGetChild(1);

			Linear constraint = new Linear();
			LinearForm leftNormalForm = new LinearForm();
			LinearForm rightNormalForm = new LinearForm();

			Double rhsConstant = visitBinaryComparison(info, lhsNode, rhsNode, constraint, leftNormalForm,
					rightNormalForm);

			if (rhsConstant == null) {
				throw new ExpressionVisitorException("Cannot use NULL in numeric comparison!");
			} else {
				if (hasVariables(leftNormalForm)) {
					if (info.inNegation) {
						addAtMostConstraint(info, constraint, rhsConstant);
					} else {
						addAtLeastConstraint(info, constraint, rhsConstant);
					}
				}
			}

			return null;
		}

		private void addNotEqualsConstraint(Info info, Linear constraint, double rhsConstant) {

			if (allVariablesBoolean(constraint) || Double.isInfinite(rhsConstant)) {

				// For boolean variables or null variables just take the negation as constraint

				if (rhsConstant == 1.0d) { // x != true
					if (info.isInOR()) {
						double upperBoundSum = calcUpperboundSum(constraint);
						for (String orVariable : info.orVariables) {
							constraint.add(-1 * upperBoundSum, orVariable);
						}
					}
					normalConstraints.add(createConstraint(constraint, Operator.LE, 0.0d));

				} else if (rhsConstant == 0.0d) { // x != false
					if (info.isInOR()) {
						double upperBoundSum = calcUpperboundSum(constraint);
						for (String orVariable : info.orVariables) {
							constraint.add(+1 * upperBoundSum, orVariable);
						}
					}
					normalConstraints.add(createConstraint(constraint, Operator.GE, 1.0d));
				} else { // x != null
					assert Double.isInfinite(rhsConstant) : "Invalid right-hand side " + rhsConstant
							+ " for 'null' comparison!";
					if (info.isInOR()) {
						double upperBoundSum = calcUpperboundSum(constraint);
						for (String orVariable : info.orVariables) {
							constraint.add(+1 * upperBoundSum, orVariable);
						}
					}
					normalConstraints.add(createConstraint(constraint, Operator.GE, Double.POSITIVE_INFINITY));
				}

			} else {

				// For integer variables

				/*
				 * A != B => A - M * OR1 <= B A + M * OR2 >= B OR1 + OR2 <= 1
				 */

				double upperBoundSum = calcUpperboundSum(constraint);
				Linear secondConstraint = cloneConstraint(constraint);

				if (info.isInOR()) {
					for (String orVariable : info.orVariables) {
						constraint.add(+1 * upperBoundSum, orVariable);
					}
					for (String orVariable : info.orVariables) {
						secondConstraint.add(-1 * upperBoundSum, orVariable);
					}
				}

				String firstOR = createORVariable();
				String secondOR = createORVariable();

				constraint.add(+1 * upperBoundSum, firstOR);
				normalConstraints
						.add(createConstraint(constraint, Operator.GE, rhsConstant + calculateEpsilon(constraint)));

				secondConstraint.add(-1 * upperBoundSum, secondOR);
				normalConstraints.add(
						createConstraint(secondConstraint, Operator.LE, rhsConstant - calculateEpsilon(constraint)));

				addConstraintForOR(ImmutableList.of(firstOR, secondOR));
			}

		}

		private void addEqualsConstraint(Info info, Linear constraint, Double rhsConstant) {

			double upperBoundSum = 0;
			if (info.isInOR()) {
				upperBoundSum = calcUpperboundSum(constraint);
			}

			if (allVariablesBoolean(constraint) || Double.isInfinite(rhsConstant)) {

				double orCoefficient;
				Operator operator;

				if (rhsConstant == 1.0d) { // x == true
					orCoefficient = +1 * upperBoundSum;
					operator = Operator.GE;
				} else { // x == false || x == null
					assert Double.isInfinite(rhsConstant) || rhsConstant == 0.0 : "Invalid right-hand side "
							+ rhsConstant + " for boolean comparison!";
					orCoefficient = -1 * upperBoundSum;
					operator = Operator.LE;
				}

				if (info.isInOR()) {
					for (String orVariable : info.orVariables) {
						constraint.add(orCoefficient, orVariable);
					}
				}

				normalConstraints.add(createConstraint(constraint, operator, rhsConstant));

			} else {

				Linear secondConstraint = cloneConstraint(constraint);
				if (info.isInOR()) {
					for (String orVariable : info.orVariables) {
						constraint.add(+1 * upperBoundSum, orVariable);
					}
					for (String orVariable : info.orVariables) {
						secondConstraint.add(-1 * upperBoundSum, orVariable);
					}
				}
				normalConstraints.add(createConstraint(constraint, Operator.GE, rhsConstant));
				normalConstraints.add(createConstraint(secondConstraint, Operator.LE, rhsConstant));
			}

		}

		private void addGreaterThanConstraint(Info info, Linear constraint, double rhsConstant) {
			if (info.isInOR()) {
				double upperBoundSum = calcUpperboundSum(constraint);
				for (String orVariable : info.orVariables) {
					constraint.add(+1 * upperBoundSum, orVariable);
				}
			}
			normalConstraints
					.add(createConstraint(constraint, Operator.GE, rhsConstant + calculateEpsilon(constraint)));
		}

		private void addLessThanConstraint(Info info, Linear constraint, double rhsConstant) {
			if (info.isInOR()) {
				double upperBoundSum = calcUpperboundSum(constraint);
				for (String orVariable : info.orVariables) {
					constraint.add(-1 * upperBoundSum, orVariable);
				}
			}
			normalConstraints
					.add(createConstraint(constraint, Operator.LE, rhsConstant - calculateEpsilon(constraint)));
		}

		private void addAtLeastConstraint(Info info, Linear constraint, double rhsConstant) {
			if (info.isInOR()) {
				double upperBoundSum = calcUpperboundSum(constraint);

				for (String orVariable : info.orVariables) {
					constraint.add(+1 * upperBoundSum, orVariable);
				}
			}
			normalConstraints.add(createConstraint(constraint, Operator.GE, rhsConstant));
		}

		private void addAtMostConstraint(Info info, Linear constraint, double rhsConstant) {
			if (info.isInOR()) {
				double upperBoundSum = calcUpperboundSum(constraint);

				for (String orVariable : info.orVariables) {
					constraint.add(-1 * upperBoundSum, orVariable);
				}
			}
			normalConstraints.add(createConstraint(constraint, Operator.LE, rhsConstant));
		}

		private Constraint createConstraint(final Linear lhs, final Operator operator, Double rhs) {
			return new Constraint("", lhs, operator, rhs);
		}

		private double calculateEpsilon(Linear constraint) {
			boolean isInteger = allVariableInteger(constraint);
			if (isInteger) {
				return 1.0d;
			} else {
				return MILPProblemBuilder.RESULT_EPSILON;
			}
		}

		private boolean allVariableInteger(Linear constraint) {
			boolean isInteger = true;
			for (Object variable : constraint.getVariables()) {
				isInteger &= isVariableRepresentedByInteger((String) variable);
			}
			return isInteger;
		}

		private boolean allVariablesBoolean(Linear constraint) {
			boolean isInteger = true;
			for (Object variable : constraint.getVariables()) {
				isInteger &= isVariableBoolean((String) variable);
			}
			return isInteger;
		}

	}

	private final static class NormalizationVisitor extends ExpressionParserDefaultVisitor {

		private Node rewriteBooleanToEquals(Node node) {
			ExprEqual newNode = new ExprEqual(ExpressionParserTreeConstants.JJTEQUAL);
			newNode.jjtAddChild(node, 0);
			ExprLitBoolean trueLiteral = new ExprLitBoolean(ExpressionParserTreeConstants.JJTLITBOOLEAN);
			trueLiteral.jjtSetValue("true");
			newNode.jjtAddChild(trueLiteral, 1);
			return newNode;
		}

		private boolean isBooleanNode(Node node) {
			return node instanceof ExprLitBoolean || node instanceof ExprVariable;
		}

		public Object visit(ExprRoot node, Object data) throws ExpressionVisitorException {
			ExprRoot root = new ExprRoot(ExpressionParserTreeConstants.JJTROOT);
			Node child = (Node) node.jjtGetChild(0).jjtAccept(this, data);
			if (isBooleanNode(child)) {
				child = rewriteBooleanToEquals(child);
			}
			root.jjtAddChild(child, 0);
			return root;
		}

		public Object visit(ExprOr node, Object data) throws ExpressionVisitorException {
			ExprOr newNode = new ExprOr(ExpressionParserTreeConstants.JJTOR);
			Node child1 = (Node) node.jjtGetChild(0).jjtAccept(this, data);
			if (isBooleanNode(child1)) {
				child1 = rewriteBooleanToEquals(child1);
			}
			newNode.jjtAddChild(child1, 0);
			Node child2 = (Node) node.jjtGetChild(1).jjtAccept(this, data);
			if (isBooleanNode(child2)) {
				child2 = rewriteBooleanToEquals(child2);
			}
			newNode.jjtAddChild(child2, 1);
			return newNode;
		}

		public Object visit(ExprAnd node, Object data) throws ExpressionVisitorException {
			ExprAnd newNode = new ExprAnd(ExpressionParserTreeConstants.JJTAND);
			Node child1 = (Node) node.jjtGetChild(0).jjtAccept(this, data);
			if (isBooleanNode(child1)) {
				child1 = rewriteBooleanToEquals(child1);
			}
			newNode.jjtAddChild(child1, 0);
			Node child2 = (Node) node.jjtGetChild(1).jjtAccept(this, data);
			if (isBooleanNode(child2)) {
				child2 = rewriteBooleanToEquals(child2);
			}
			newNode.jjtAddChild(child2, 1);
			return newNode;
		}

		public Object visit(ExprNot node, Object data) throws ExpressionVisitorException {
			ExprNot newNode = new ExprNot(ExpressionParserTreeConstants.JJTNOT);
			Node child = (Node) node.jjtGetChild(0).jjtAccept(this, data);
			if (isBooleanNode(child)) {
				child = rewriteBooleanToEquals(child);
			}
			newNode.jjtAddChild(child, 0);
			return newNode;
		}

		public Object visit(ExprEqual node, Object data) throws ExpressionVisitorException {
			ExprEqual newNode = new ExprEqual(ExpressionParserTreeConstants.JJTEQUAL);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprNotEqual node, Object data) throws ExpressionVisitorException {
			ExprNotEqual newNode = new ExprNotEqual(ExpressionParserTreeConstants.JJTNOTEQUAL);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprLessThan node, Object data) throws ExpressionVisitorException {
			ExprLessThan newNode = new ExprLessThan(ExpressionParserTreeConstants.JJTLESSTHAN);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprAtMost node, Object data) throws ExpressionVisitorException {
			ExprAtMost newNode = new ExprAtMost(ExpressionParserTreeConstants.JJTATMOST);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprGreaterThan node, Object data) throws ExpressionVisitorException {
			ExprGreaterThan newNode = new ExprGreaterThan(ExpressionParserTreeConstants.JJTGREATERTHAN);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprAtLeast node, Object data) throws ExpressionVisitorException {
			ExprAtLeast newNode = new ExprAtLeast(ExpressionParserTreeConstants.JJTATLEAST);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprPlus node, Object data) throws ExpressionVisitorException {
			ExprPlus newNode = new ExprPlus(ExpressionParserTreeConstants.JJTPLUS);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprMinus node, Object data) throws ExpressionVisitorException {
			ExprMinus newNode = new ExprMinus(ExpressionParserTreeConstants.JJTMINUS);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprMult node, Object data) throws ExpressionVisitorException {
			ExprMult newNode = new ExprMult(ExpressionParserTreeConstants.JJTMULT);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprDiv node, Object data) throws ExpressionVisitorException {
			ExprDiv newNode = new ExprDiv(ExpressionParserTreeConstants.JJTDIV);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprNegation node, Object data) throws ExpressionVisitorException {
			ExprNegation newNode = new ExprNegation(ExpressionParserTreeConstants.JJTNEGATION);
			newNode.jjtSetValue(node.jjtGetValue());
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			return newNode;
		}

		public Object visit(ExprVariable node, Object data) throws ExpressionVisitorException {
			ExprVariable newNode = new ExprVariable(ExpressionParserTreeConstants.JJTVARIABLE);
			newNode.jjtSetValue(node.jjtGetValue());
			return newNode;
		}

		public Object visit(ExprLitInteger node, Object data) throws ExpressionVisitorException {
			ExprLitInteger newNode = new ExprLitInteger(ExpressionParserTreeConstants.JJTLITINTEGER);
			newNode.jjtSetValue(node.jjtGetValue());
			return newNode;
		}

		public Object visit(ExprLitDouble node, Object data) throws ExpressionVisitorException {
			ExprLitDouble newNode = new ExprLitDouble(ExpressionParserTreeConstants.JJTLITDOUBLE);
			newNode.jjtSetValue(node.jjtGetValue());
			return newNode;
		}

		public Object visit(ExprLitString node, Object data) throws ExpressionVisitorException {
			ExprLitString newNode = new ExprLitString(ExpressionParserTreeConstants.JJTLITSTRING);
			newNode.jjtSetValue(node.jjtGetValue());
			return newNode;
		}

		public Object visit(ExprLitBoolean node, Object data) throws ExpressionVisitorException {
			ExprLitBoolean newNode = new ExprLitBoolean(ExpressionParserTreeConstants.JJTLITBOOLEAN);
			newNode.jjtSetValue(node.jjtGetValue());
			return newNode;
		}

		public Object visit(ExprLitNull node, Object data) throws ExpressionVisitorException {
			ExprLitNull newNode = new ExprLitNull(ExpressionParserTreeConstants.JJTLITNULL);
			newNode.jjtSetValue(node.jjtGetValue());
			return newNode;
		}

		public Object visit(ExprFunction node, Object data) throws ExpressionVisitorException {
			throw new ExpressionVisitorException("Cannot normalize expressions containing functions!");
		}

	}

	private static final double MAX_VALUE = 999999999.0d; // Should not be too big, otherwise introduces errors in lp_solve
	private static final String OR_PREFIX = "O%";

	private int numBooleanVar = 0;

	private final String booleanVarPrefix;

	private final List<String> booleanVariables;
	private final List<Constraint> extraOrConstraints;
	private final List<Constraint> normalConstraints;

	private final StringDiscretizer discretizer;
	private final UnassignedMode unassignedMode;

	private final Map<String, Variable> variables;

	public IlpExpressionTemplate(StringDiscretizer discretizer, Map<String, Variable> variables,
			UnassignedMode unassignedMode, GuardExpression expression, String booleanVarPrefix)
			throws NotLinearException {
		this.discretizer = discretizer;
		this.variables = variables;
		this.unassignedMode = unassignedMode;
		this.booleanVarPrefix = booleanVarPrefix;

		this.booleanVariables = new ArrayList<>();
		this.extraOrConstraints = new ArrayList<>();
		this.normalConstraints = new ArrayList<>();

		try {
			expression = normalizeExpression(expression);
			expression.visit(new ExpressionConversionVisitor(), null);
		} catch (NotLinearException e) {
			throw e;
		} catch (ExpressionVisitorException e) {
			throw new RuntimeException("Error parsing guard expression " + expression, e);
		}
	}

	private static GuardExpression normalizeExpression(GuardExpression expression) throws ExpressionVisitorException {
		ExprRoot exprRoot = (ExprRoot) expression.visit(new NormalizationVisitor(), null);
		return GuardExpression.Factory.newInstance(exprRoot);
	}

	public void addToILPProblem(final ILPProblem ilpProblem, final TObjectIntMap<String> timesWritten,
			final Set<String> variablesToWrite, final String uniqueSuffix) {
		for (int j = 0; j < extraOrConstraints.size(); j++) {
			Constraint c = extraOrConstraints.get(j);
			Linear linear = createOrLinear(uniqueSuffix, c);
			ilpProblem.addORConstraints(linear, c.getOperator(), c.getRhs());
		}
		for (int j = 0; j < booleanVariables.size(); j++) {
			String varName = booleanVariables.get(j);
			ilpProblem.setVarType(varName.concat(uniqueSuffix), Boolean.class);
		}
		for (int j = 0; j < normalConstraints.size(); j++) {
			Constraint c = normalConstraints.get(j);
			addNormalConstraint(ilpProblem, timesWritten, variablesToWrite, uniqueSuffix, c);
		}
	}

	private void addNormalConstraint(final ILPProblem ilpProblem, final TObjectIntMap<String> timesWritten,
			final Set<String> variablesToWrite, final String uniqueSuffix, Constraint templateConstraint) {
		Linear linear = new Linear();
		Linear lhs = templateConstraint.getLhs();
		Double rhs = (Double) templateConstraint.getRhs();
		boolean isNullComparison = rhs.isInfinite();

		String variableName = null;

		for (int k = 0; k < lhs.size(); k++) {
			Term term = lhs.get(k);
			String varName = getAsVariable(term);
			if (varName.startsWith(OR_PREFIX)) {
				// OR term
				Number coefficient = term.getCoefficient();
				if (isNullComparison) {
					// null comparison has a fixed coefficient of 1.0
					coefficient = 1.0d;
				}
				Term orTerm = new Term(varName.concat(uniqueSuffix), coefficient);
				linear.add(orTerm);
			} else {
				variableName = convertVariableName(varName, timesWritten, variablesToWrite);
				if (isNullComparison) {
					// NULL term
					if (DataStateHelper.isPrimeVar(varName)) {
						throw new IllegalArgumentException("Does not support null comparisons with prime variables!");
					}
					if (unassignedMode != UnassignedMode.NULL) {
						// just ignore this constraint
						return;
					} // otherwise handle the situation later
				} else {
					// normal term					
					linear.add(new Term(variableName, term.getCoefficient()));
				}
			}
		}

		if (variableName == null) {
			throw new IllegalArgumentException();
		}

		if (rhs.isInfinite()) {
			ilpProblem.addNullConstraint(variableName, linear, templateConstraint.getOperator(), rhs);
		} else {
			ilpProblem.addGuardConstraints(linear, templateConstraint.getOperator(), rhs);
		}

	}

	private static Linear createOrLinear(String uniqueSuffix, Constraint c) {
		Linear linear = new Linear();
		Linear lhs = c.getLhs();
		for (int k = 0; k < lhs.size(); k++) {
			Term term = lhs.get(k);
			String variableName = getAsVariable(term).concat(uniqueSuffix);
			linear.add(new Term(variableName, term.getCoefficient()));
		}
		return linear;
	}

	private static String getAsVariable(Term term) {
		return (String) term.getVariable();
	}

	private static String convertVariableName(final String varName, final TObjectIntMap<String> timesWritten,
			final Set<String> listVariableToWrite) {
		if (DataStateHelper.isPrimeVar(varName)) {
			String processVar = DataStateHelper.stripPrime(varName);
			int times = timesWritten.get(processVar);
			if (times == 0) {
				throw new IllegalStateException("There should have been at least one write operation!");
			}
			return transformToIlpVarName(processVar, times);
		} else {					
			// add one to get the previous variable instead of the current prime variable 
			// (assumes control flow steps are iterated through backwards!)
			// if none is written this leads to: 1
			return transformToIlpVarName(varName, timesWritten.get(varName) + 1);
		}
	}


	private final String createORVariable() {
		// Only a single thread of execution!
		int currentBooleanVar = numBooleanVar++;
		return OR_PREFIX.concat(booleanVarPrefix).concat(String.valueOf(currentBooleanVar));
	}

	private final double calcUpperBound(final String varName) {
		double upperbound;
		if (getVariableType(varName) == Boolean.class) {
			upperbound = 1;
		} else if (getUpperBound(varName) != null) {
			upperbound = getUpperBound(varName);
		} else {
			upperbound = MAX_VALUE;
		}
		return upperbound;
	}

	private Double getUpperBound(String varName) {
		Variable variable = variables.get(varName);
		if (variable != null) {
			return variable.getIlpUpperBound();
		}
		return null;
	}

	private Class<?> getVariableType(String varName) {
		Variable variable = variables.get(varName);
		if (variable != null) {
			return variable.getType();
		}
		return null;
	}

	private final boolean isVariableRepresentedByInteger(final String varName) {
		String realName = getRealVarName(varName);
		Class<?> type = getVariableType(realName);
		if (type == null) {
			return true; // it is an OR variable which is boolean
		} else {
			return type == Integer.class || type == Long.class || type == String.class || type == Boolean.class;
		}
	}

	private final boolean isVariableBoolean(final String varName) {
		String realName = getRealVarName(varName);
		Class<?> type = getVariableType(realName);
		if (type == null) {
			return true; // it is an OR variable which is boolean
		} else {
			return type == Boolean.class;
		}
	}

	private static String getRealVarName(final String var) {
		if (DataStateHelper.isPrimeVar(var)) {
			return DataStateHelper.stripPrime(var);
		}
		return var;
	}

	private double calcUpperboundSum(Linear constraint) {
		double upperBoundSum = 0.0d;
		for (Object constraintVar : constraint.getVariables()) {
			upperBoundSum += calcUpperBound(getRealVarName((String) constraintVar));
		}
		upperBoundSum = Math.min(MAX_VALUE, upperBoundSum);
		return upperBoundSum;
	}

	private static Linear cloneConstraint(final Linear constraint) {
		Linear retValue = new Linear();
		for (int i = 0; i < constraint.size(); i++) {
			Term term = constraint.get(i);
			retValue.add(term.getCoefficient().doubleValue(), term.getVariable());
		}
		return retValue;
	}

	private static String transformToIlpVarName(final String var, final int times) {
		return var.concat(String.valueOf(times));
	}
}
