package org.processmining.plugins.balancedconformance.dataflow;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.Set;

import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.DataConformance.DataAlignment.GenericTrace;
import org.processmining.plugins.DataConformance.DataAlignment.SingleTraceDataAlignmentBuilder;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.framework.ExecutionTrace;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration.UnassignedMode;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult.Step;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataState;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateFactory;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateHelper;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;
import org.processmining.plugins.balancedconformance.dataflow.exception.InfeasibleMILPProblem;
import org.processmining.plugins.balancedconformance.dataflow.exception.MILPProblemException;
import org.processmining.plugins.balancedconformance.dataflow.lp.ILPProblem;
import org.processmining.plugins.balancedconformance.dataflow.lp.ILPProblemImpl;
import org.processmining.plugins.balancedconformance.dataflow.util.ExpressionEvaluator;
import org.processmining.plugins.balancedconformance.functions.VirtualVariable;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizer;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;
import org.processmining.plugins.balancedconformance.mapping.Variable;
import org.processmining.plugins.balancedconformance.mapping.VariableAwareWriteOperationCost;
import org.processmining.plugins.petrinet.replayresult.StepTypes;

import com.google.common.base.Throwables;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.math.DoubleMath;

import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.hash.TObjectIntHashMap;
import net.sf.javailp.Linear;
import net.sf.javailp.Operator;
import net.sf.javailp.Result;
import net.sf.javailp.Solver;
import net.sf.javailp.SolverFactory;

/**
 * Rewritten, optimized variant of the {@link SingleTraceDataAlignmentBuilder}
 * from the DataAwareReplayer package. Supports partial alignments and has an
 * improved runtime and memory footprint.
 * 
 * @author F. Mannhardt
 * 
 */
public final class MILPProblemBuilderImpl implements MILPProblemBuilder {

	private enum CalculationMode {
		FULL, PARTIAL
	}

	private final static class MILPVariable {

		private Variable variable;
		private String name;
		private String booleanName;
		private String defaultValueBooleanName;

		private Class<?> type;
		private Double upperBound;
		private Double lowerBound;
		private Double defaultValue;

		public static MILPVariable newVariable(
				Entry<String, org.processmining.plugins.balancedconformance.mapping.Variable> entry) {
			MILPVariable variable = new MILPVariable();
			String varName = entry.getKey();
			variable.setVariable(entry.getValue());
			variable.setName(varName);
			variable.setBooleanVarName(varName.concat("_D"));
			variable.setDefaultValueBooleanName(varName.concat("_INITIAL"));
			org.processmining.plugins.balancedconformance.mapping.Variable logVariable = entry.getValue();
			variable.setType(logVariable.getType());
			Double upperBound = logVariable.getIlpUpperBound();
			if (upperBound != null) {
				variable.setInternalUpperBound(logVariable.getIlpUpperBound());
			}
			Double lowerBound = logVariable.getIlpLowerBound();
			if (lowerBound != null) {
				variable.setInternalLowerBound(logVariable.getIlpLowerBound());
			}
			variable.defaultValue = logVariable.getIlpDefaultValue();
			return variable;
		}

		public String getInternalName(int variableCount) {
			return name.concat(String.valueOf(variableCount));
		}

		public Class<?> getInternalType() {
			if (getType() == String.class)
				return Integer.class;
			if (getType() == Float.class)
				return Double.class;
			if (getType() == java.util.Date.class || getType() == Long.class)
				return Integer.class;
			return getType();
		}

		public Class<?> getType() {
			return type;
		}

		public void setType(Class<?> type) {
			this.type = type;
		}

		public String getName() {
			return name;
		}

		public void setName(String name) {
			this.name = name;
		}

		public Double getUpperBound() {
			return upperBound;
		}

		public void setInternalUpperBound(Double upperBound) {
			this.upperBound = upperBound;
		}

		public Double getLowerBound() {
			return lowerBound;
		}

		public void setInternalLowerBound(Double lowerBound) {
			this.lowerBound = lowerBound;
		}

		public String getBooleanVarName(int variableCount) {
			return booleanName.concat(String.valueOf(variableCount));
		}

		public void setBooleanVarName(String booleanName) {
			this.booleanName = booleanName;
		}

		public String getDefaultValueBooleanName() {
			return defaultValueBooleanName;
		}

		public void setDefaultValueBooleanName(String defaultValueBooleanName) {
			this.defaultValueBooleanName = defaultValueBooleanName;
		}

		public Variable getVariable() {
			return variable;
		}

		public void setVariable(Variable variable) {
			this.variable = variable;
		}

		public String toString() {
			return String.format("Variable %s (BoolVar: %s, NullVar: %s, InitialVar: %s, Type: %s, Bounds: (%s, %s)",
					name, booleanName, defaultValueBooleanName, type, upperBound, lowerBound);
		}

	}

	public final static class ScopedVariable {

		MILPVariable variable;
		int scope;

		public ScopedVariable(MILPVariable variable, int scope) {
			super();
			this.variable = variable;
			this.scope = scope;
		}

		public String getIlpName() {
			return variable.getInternalName(scope);
		}

		@Override
		public String toString() {
			return String.format("%s", variable.getInternalName(scope));
		}

	}

	private static boolean DEBUG = false;

	private final SolverFactory factory;
	private final DataStateFactory dataStateFactory;
	private final LogMapping mapping;

	private final double aStarDelta;
	private final boolean includeVirtualVariablesInTraces;

	private Map<String, MILPVariable> variables;

	private boolean usePartialProblems = true;

	static {
		System.loadLibrary("lpsolve55");
		System.loadLibrary("lpsolve55j");
	}

	public MILPProblemBuilderImpl(LogMapping mapping, SolverFactory solverFactory, DataStateFactory dataStateFactory,
			double aStarDelta, boolean includeVirtualVariablesInTraces, boolean usePartialProblems) {
		this.mapping = mapping;
		this.factory = solverFactory;
		this.dataStateFactory = dataStateFactory;
		this.aStarDelta = aStarDelta;
		this.includeVirtualVariablesInTraces = includeVirtualVariablesInTraces;
		this.usePartialProblems = usePartialProblems;
		this.variables = Maps.newHashMapWithExpectedSize(mapping.getVariables().size());
		for (Entry<String, org.processmining.plugins.balancedconformance.mapping.Variable> entry : mapping
				.getVariables().entrySet()) {
			variables.put(entry.getKey(), MILPVariable.newVariable(entry));
		}
	}

	public FullDataAlignedTrace createAlignment(TraceMapping mapping, ControlFlowAlignmentResult controlFlowAlignment)
			throws MILPProblemException {

		List<ScopedVariable[]> variablesByStep = new ArrayList<>(controlFlowAlignment.getApproximateSize());
		ImmutableList<Step> stepList = ImmutableList.copyOf(controlFlowAlignment.getSteps());

		double ilpCost = 0.0d;
		Result result = null;

		ILPProblem ilpProblem = new ILPProblemImpl();
		TObjectIntMap<String> timesWrittenVariable = new TObjectIntHashMap<>(variables.size(), 0.25f, 0);
		double additionalCost = addGuardsAndAssignments(mapping, stepList, controlFlowAlignment.getLastRecord(),
				timesWrittenVariable, ilpProblem, variablesByStep, CalculationMode.FULL);

		if (Double.isInfinite(additionalCost)) {
			return FullDataAlignedTrace.IMPOSSIBLE;
		}

		addDefaultOrFreeVariablesToILP(timesWrittenVariable, ilpProblem);

		Solver solver = getSolver();
		result = ilpProblem.solve(solver);

		if (result == null && ilpProblem.getObjectiveSize() == 0 && ilpProblem.getConstraintsCount() == 0) {
			// No ILP problem to solve, but maybe some variables are not written
			ilpCost = additionalCost;
		} else if (result == null) {
			// Infeasible ILP problem OR timeout - try again
			result = ilpProblem.solve(solver);
			if (result == null) {

				assert !areAllGuardsTrue(mapping, controlFlowAlignment, variablesByStep, ilpProblem, result,
						false) : assertErrorGuardsViolatedAssertionMessage(mapping, controlFlowAlignment, null, result);

				throw new InfeasibleMILPProblem("Alignment:\n" + controlFlowAlignment.toString() + "\n\nILP Problem:\n"
						+ ilpProblem.toString());
			} else {
				System.out.println("Solved ILP in 2nd try " + ilpProblem.toString());
				ilpCost = (float) (result.getObjective().doubleValue() + additionalCost);
			}
		} else {
			ilpCost = (float) (result.getObjective().doubleValue() + additionalCost);
		}

		assert areAllGuardsTrue(mapping, controlFlowAlignment, variablesByStep, ilpProblem, result,
				false) : assertErrorGuardsViolatedAssertionMessage(mapping, controlFlowAlignment, null, result);

		final AlignedTraces processAndLogTrace = createFullAlignedTrace(mapping, ilpProblem, result, stepList,
				variablesByStep);

		final int dataCost;
		if (ilpCost >= Integer.MAX_VALUE) {
			return FullDataAlignedTrace.IMPOSSIBLE;
		} else {
			dataCost = (int) ilpCost;
		}

		return new FullDataAlignedTrace() {

			public int getDataCost() {
				return dataCost;
			}

			public ExecutionTrace getProcessTracePrefix() {
				return processAndLogTrace.getProcessTrace();
			}

			public ExecutionTrace getLogTracePrefix() {
				return processAndLogTrace.getLogTrace();
			}

			public String toString() {
				return String.format("Cost: %s, Model: %s, Log: %s", getDataCost(), getProcessTracePrefix(),
						getLogTracePrefix());
			}

		};
	}

	public PartialDataAlignedTrace createPartialAlignment(TraceMapping traceMapping,
			ControlFlowAlignmentResult controlFlowAlignment) throws MILPProblemException {

		if (DEBUG) {
			System.out.println("\n\nAligning prefix of trace " + traceMapping.getTraceName());
		}

		ImmutableList<Step> alignmentList = ImmutableList.copyOf(controlFlowAlignment.getSteps());
		List<ScopedVariable[]> variablesByStep = new ArrayList<>(controlFlowAlignment.getApproximateSize());

		ILPProblem ilpProblem = new ILPProblemImpl();

		double ilpCost = 0.0d;
		Result result = null;

		TObjectIntMap<String> timesWrittenVariable = new TObjectIntHashMap<>(variables.size(), 0.5f, 0);

		double additionalCost = addGuardsAndAssignments(traceMapping, alignmentList,
				controlFlowAlignment.getLastRecord(), timesWrittenVariable, ilpProblem, variablesByStep,
				usePartialProblems ? CalculationMode.PARTIAL : CalculationMode.FULL);

		if (Double.isInfinite(additionalCost)) {
			return PartialDataAlignedTrace.IMPOSSIBLE;
		}

		addDefaultOrFreeVariablesToILP(timesWrittenVariable, ilpProblem);

		Solver solver = getSolver();
		result = ilpProblem.solve(solver);

		assert reCheckILP(ilpProblem, result, solver) : "Two rounds of ILP solving disagree with each other!";

		if (result == null && ilpProblem.getObjectiveSize() == 0 && ilpProblem.getConstraintsCount() == 0) {
			// No ILP problem to solve, but maybe some variables are not written
			ilpCost = additionalCost;
			if (DEBUG) {
				System.out.println("No ILP needed for prefix " + controlFlowAlignment);
			}
		} else if (result == null) {
			// Unfeasible ILP problem OR timeout
			if (DEBUG) {
				System.out.println("Infeasible ILP for prefix " + controlFlowAlignment);
				assert testAgainWithFullAlignment(traceMapping,
						controlFlowAlignment) : assertErrorGuardsNotViolatedAssertionMessage(traceMapping,
								controlFlowAlignment, ilpProblem, result);
			}
			return PartialDataAlignedTrace.IMPOSSIBLE;
		} else {
			ilpCost = (float) (result.getObjective().doubleValue() + additionalCost);
			if (DEBUG) {
				System.out.println("Solving ILP for prefix " + controlFlowAlignment);
				System.out.println(ilpProblem);
				System.out.println(result);
			}
		}

		if (ilpCost >= Integer.MAX_VALUE) {
			return PartialDataAlignedTrace.IMPOSSIBLE;
		}

		assert lastGuardTrue(traceMapping, controlFlowAlignment, variablesByStep, ilpProblem, result,
				false) : assertErrorGuardsViolatedAssertionMessage(traceMapping, controlFlowAlignment, ilpProblem,
						result);

		PartialDataAlignedTrace partialAlignedTrace = createPartialAlignedTrace(traceMapping, ilpProblem, result,
				ilpCost, controlFlowAlignment, alignmentList, variablesByStep);

		// compare to full alignment w/o delta scaling
		assert createAlignment(traceMapping, controlFlowAlignment)
				.getDataCost() == (int) (partialAlignedTrace.getDataCost()
						/ aStarDelta) : assertErrorFullDataAlignmentDisagrees(traceMapping, controlFlowAlignment,
								ilpProblem, result, (int) (partialAlignedTrace.getDataCost() / aStarDelta),
								partialAlignedTrace);

		return partialAlignedTrace;
	}

	private boolean testAgainWithFullAlignment(TraceMapping mapping, ControlFlowAlignmentResult controlFlowAlignment)
			throws MILPProblemException {
		try {
			return createAlignment(mapping, controlFlowAlignment) == FullDataAlignedTrace.IMPOSSIBLE;
		} catch (InfeasibleMILPProblem e) {
			return true;
		}
	}

	private Solver getSolver() {
		return factory.get();
	}

	private void addDefaultOrFreeVariablesToILP(TObjectIntMap<String> timesWritten, ILPProblem ilpProblem) {
		if (mapping.getUnassignedMode() == UnassignedMode.DEFAULT) {
			// uninitialized variables are initialized with a configured 'default' value
			for (MILPVariable variable : variables.values()) {
				int variableCount = addVariableToILP(variable, ilpProblem, timesWritten);
				// Add constraint for default value to ILP problem
				Double defaultValue = variable.defaultValue;
				if (defaultValue != null) {
					String ilpVarName = variable.getInternalName(variableCount);
					addVariableAssignment(ilpProblem, ilpVarName, variable.getDefaultValueBooleanName(),
							defaultValue.doubleValue());
					// Add maximum cost for violation of default value as default value is not a 'normal' write operation. It cannot be 'wrong', and a violation leads to a unsound DPN-net!
					ilpProblem.addObjective(Integer.valueOf(Integer.MAX_VALUE), variable.getDefaultValueBooleanName());
				} else {
					// No default value, it is treated as free
					addVariableToILP(variable, ilpProblem, timesWritten);
				}
			}
		} else if (mapping.getUnassignedMode() == UnassignedMode.FREE) {
			// all variables are free, i.e., uninitialized variables can be initialized with any value
			for (MILPVariable variable : variables.values()) {
				addVariableToILP(variable, ilpProblem, timesWritten);
			}
		} // else handle explicit comparisons with NULL in the ILPTemplateExpression

	}

	private double addGuardsAndAssignments(TraceMapping traceMapping, List<Step> alignment, DRecord lastRecord,
			TObjectIntMap<String> timesWritten, ILPProblem problem, List<ScopedVariable[]> variablesByStep,
			CalculationMode calculationMode) {

		double additionalCost = 0.0;
		Set<String> requiredVariablesFromPrefix = new HashSet<>();

		// Add real process trace
		for (int stepNr = alignment.size() - 1; stepNr >= 0; stepNr--) {
			Step step = alignment.get(stepNr);
			if (step.getType() != StepTypes.L) {
				additionalCost = handleProcessStep(traceMapping, lastRecord, timesWritten, problem, variablesByStep,
						additionalCost, requiredVariablesFromPrefix, stepNr, step);
			}
			if (calculationMode == CalculationMode.PARTIAL && requiredVariablesFromPrefix.isEmpty()) {
				break; // no need to go further back
			}
		}
		return handleNullVariables(problem, additionalCost, requiredVariablesFromPrefix, timesWritten);
	}

	private double handleProcessStep(TraceMapping traceMapping, DRecord lastRecord, TObjectIntMap<String> timesWritten,
			ILPProblem problem, List<ScopedVariable[]> variablesByStep, double additionalCost,
			Set<String> requiredVariablesFromPrefix, int stepNr, Step step) {
		Set<String> listVariableToWrite = mapping.getVariablesToWrite().get(step.getModelNode());

		// We provide values for these variables
		requiredVariablesFromPrefix.removeAll(listVariableToWrite);

		additionalCost = handleAssignmentsAndGuards(traceMapping, lastRecord, timesWritten, problem, variablesByStep,
				additionalCost, stepNr, step, listVariableToWrite);

		// We need values for these variables from the prefix
		requiredVariablesFromPrefix.addAll(getNormalVariablesForGuard(step));
		return additionalCost;
	}

	private double handleAssignmentsAndGuards(TraceMapping traceMapping, DRecord lastRecord,
			TObjectIntMap<String> timesWritten, ILPProblem problem, List<ScopedVariable[]> variablesByStep,
			double additionalCost, int stepNr, Step step, Set<String> listVariableToWrite) {
		// First add variables (side effect on 'timesWritten'!)
		ScopedVariable[] mappingVariables = addVariables(step.getModelNode(), timesWritten, problem, variablesByStep,
				listVariableToWrite);

		// Then, add guards (they might use the just added variable)
		if (hasRelevantGuard(step)) {
			addGuards(stepNr, timesWritten, problem, step, listVariableToWrite);
		}
		if (hasVariableAssignments(mappingVariables)) {
			additionalCost += addAllVariableAssignments(traceMapping, lastRecord, problem, step, mappingVariables,
					mapping.getVariableCost());

		}
		return additionalCost;
	}

	private double handleNullVariables(ILPProblem problem, double additionalCost,
			Set<String> requiredVariablesFromPrefix, TObjectIntMap<String> timesWritten) {

		if (!requiredVariablesFromPrefix.isEmpty()) {
			// there is a missing variable
			for (Iterator<String> iterator = requiredVariablesFromPrefix.iterator(); iterator.hasNext();) {
				String variable = iterator.next();
				int nextIndex = timesWritten.get(variable) + 1; // this would be the required index
				ScopedVariable scopedVariable = new ScopedVariable(variables.get(variable), nextIndex);
				String ilpVariable = scopedVariable.getIlpName();
				if (problem.hasNullConstraint(ilpVariable)) {
					// Record the fact that this variable was NULL beforehand
					additionalCost += problem.writeNullConstraint(ilpVariable, true);
					iterator.remove();
				}
			}
		}

		if (!requiredVariablesFromPrefix.isEmpty() && mapping.getUnassignedMode() == UnassignedMode.NULL) {
			// alignment is impossible since a variable is NULL
			//			throw new InfeasibleMILPProblem("Alignment impossible. Unassigned variables are used in a guard expression "
			//					+ requiredVariablesFromPrefix + ". Check whether variables can be written before they are used.");
			// return infinity instead of error, there might be another path that is possible!
			return Double.POSITIVE_INFINITY;
		}
		return additionalCost;
	}

	private static boolean hasVariableAssignments(ScopedVariable[] mappingVariables) {
		return mappingVariables.length > 0;
	}

	private void addGuards(int stepNr, TObjectIntMap<String> timesWritten, ILPProblem ilpProblem, Step step,
			Set<String> listVariableToWrite) {
		IlpExpressionTemplate expressionTemplate = mapping.getTransitionNameToExpression().get(step.getModelNode());
		expressionTemplate.addToILPProblem(ilpProblem, timesWritten, listVariableToWrite, String.valueOf(stepNr));
	}

	private ScopedVariable[] addVariables(Transition transition, TObjectIntMap<String> timesWritten,
			ILPProblem ilpProblem, List<ScopedVariable[]> variablesByStep, Set<String> listVariableToWrite) {
		int j = 0;
		ScopedVariable[] mappingVariables = new ScopedVariable[listVariableToWrite.size()];
		for (String var : listVariableToWrite) {
			MILPVariable variable = variables.get(var);
			String activity = mapping.getTransitionToName().get(transition);
			boolean isVariableFinal = mapping.getVariableCost().isFinal(activity, variable.getVariable());
			if (variable != null && !isVariableFinal) {
				int variableCount = addVariableToILP(variable, ilpProblem, timesWritten);
				mappingVariables[j] = new ScopedVariable(variable, variableCount);
				j++;
			}
		}
		if (j != mappingVariables.length) {
			mappingVariables = Arrays.copyOf(mappingVariables, j);
		}
		variablesByStep.add(mappingVariables);
		return mappingVariables;
	}

	private boolean hasRelevantGuard(Step step) {
		Transition transition = step.getModelNode();
		if (transition != null && mapping.getGuardExpression(step.getModelNode()) != null // has guard
				&& !mapping.finalVariablesInGuard(transition)) { // does not contain final variables
			return true;
		}
		return false;
	}

	private Set<String> getNormalVariablesForGuard(Step step) {
		Transition transition = step.getModelNode();
		if (transition != null) {
			GuardExpression guardExpression = mapping.getGuardExpression(step.getModelNode());
			if (guardExpression != null) {
				return guardExpression.getNormalVariables();
			}
		}
		return ImmutableSet.of();
	}

	private static int addVariableToILP(MILPVariable variable, ILPProblem ilpProblem,
			TObjectIntMap<String> timesWritten) {
		int variableCount = timesWritten.adjustOrPutValue(variable.getName(), 1, 1);
		String ilpVarName = variable.getInternalName(variableCount);
		ilpProblem.setVarType(ilpVarName, variable.getInternalType());
		if (variable.getUpperBound() != null) {
			ilpProblem.setVarUpperBound(ilpVarName, variable.getUpperBound());
		}
		if (variable.getLowerBound() != null) {
			ilpProblem.setVarLowerBound(ilpVarName, variable.getLowerBound());
		}
		return variableCount;
	}

	private PartialDataAlignedTrace createPartialAlignedTrace(TraceMapping traceMapping, ILPProblem ilpProblem,
			Result result, double ilpCost, ControlFlowAlignmentResult controlFlowAlignment,
			ImmutableList<Step> alignmentList, List<ScopedVariable[]> variablesByStep) {

		DRecord previousRecord = controlFlowAlignment.getLastRecord().getPredecessor();
		int dataCost = 0;
		int previousDataState = getPreviousDataState(previousRecord);

		if (!usePartialProblems) {
			dataCost = (int) Math.ceil(ilpCost * aStarDelta); // only new costs
		} else if (previousDataState == DataStateStore.EMPTY_STATE_INDEX) {
			dataCost = (int) Math.ceil(ilpCost * aStarDelta); // only new costs
		} else if (previousRecord.getDataCost() == 0) {
			dataCost = (int) Math.ceil(ilpCost * aStarDelta); // only new costs
		} else if (alignmentList.size() > 1) {

			// There have already been violations in the data perspective
			// We need to merge those with the newly detected violations

			// Take old data state as base
			previousDataState = previousRecord.getDataState();
			// Take old costs as base
			dataCost = previousRecord.getDataCost();
			// Remove all costs for write operations that fall into our scope (i.e., so long as we have ScopedVariables
			Iterator<ScopedVariable[]> varResultIter = variablesByStep.iterator();
			if (varResultIter.hasNext()) {
				varResultIter.next();
			}

			// start from previous step as the current one cannot be part of the old cost
			for (int stepNr = alignmentList.size() - 2; stepNr >= 0; stepNr--) {
				if (!varResultIter.hasNext()) {
					break; // end of our scope
				}

				Step step = alignmentList.get(stepNr);
				// Remove cost for each write operation in the old result that was in our scope 
				if (step.getModelNode() != null) {
					varResultIter.next();

					DRecord record = step.getRecordForStep();
					// cost delta caused by a guard in our scope -> remove
					int costDelta = getDataCostDelta(record);
					// cost must be because of an incorrect write operation or missing write operation in our scope
					// so subtract from old cost 
					dataCost -= costDelta;
				}
			}

			// Add new costs
			dataCost += (int) Math.ceil(ilpCost * aStarDelta);
		}

		DataState dataState = createDataStateFromILP(traceMapping, ilpProblem, result, controlFlowAlignment,
				variablesByStep, previousDataState);
		return new PartialDataAlignedTrace(dataCost, dataState);
	}

	private int getPreviousDataState(DRecord previousRecord) {
		if (previousRecord == null) {
			return DataStateStore.EMPTY_STATE_INDEX;
		} else {
			return previousRecord.getDataState();
		}
	}

	/**
	 * Data cost added by record.
	 * 
	 * @param record
	 * @return
	 */
	private int getDataCostDelta(DRecord record) {
		if (record.getPredecessor() != null) {
			DRecord prevRecord = record.getPredecessor();
			return record.getDataCost() - prevRecord.getDataCost();
		} else {
			return record.getDataCost();
		}
	}

	private DataState createDataStateFromILP(TraceMapping traceMapping, ILPProblem ilpProblem, Result result,
			ControlFlowAlignmentResult controlFlowAlignment, List<ScopedVariable[]> variablesByStep,
			int previousDataState) {
		// Get last DataState (which was correct for the previous step) 
		DataState variableAssignments;
		if (previousDataState == DataStateStore.EMPTY_STATE_INDEX) {
			variableAssignments = dataStateFactory.newDataState();
		} else {
			variableAssignments = traceMapping.getDataStateStore().get(previousDataState).clone(true);
		}

		// Iterate backwards so the last step is the actual last step
		for (ListIterator<ScopedVariable[]> iterator = variablesByStep.listIterator(variablesByStep.size()); iterator
				.hasPrevious();) {
			ScopedVariable[] scopedVariables = iterator.previous();
			boolean isLastStep = !iterator.hasPrevious();

			for (int j = 0; j < scopedVariables.length; j++) {
				ScopedVariable scopedVar = scopedVariables[j];
				Number processValue = extractVariableFromResult(ilpProblem, result, scopedVar);
				String processVariable = scopedVar.variable.getName();
				String varName = determineVariableName(isLastStep, processVariable);
				Class<?> varType = scopedVar.variable.getType();
				DataStateHelper.storeValueFromIlp(variableAssignments, varName, processValue, varType);
			}
		}

		Transition lastTransition = controlFlowAlignment.getLastTransition();
		if (lastTransition != null) {
			final Set<String> variableWritings = mapping.getVariablesToWrite().get(lastTransition);
			for (String variableName : variableWritings) {
				Variable variable = mapping.getVariables().get(variableName);
				if (mapping.getVariableCost().isFinal(mapping.getTransitionToName().get(lastTransition), variable)) {
					// variable was not part of the MILP problem, just write the value as guard needs to be true by now
					Class<?> varType = variable.getType();
					Object value = traceMapping.getVariableAssignmentFor(variableName,
							controlFlowAlignment.getLastRecord(), controlFlowAlignment.getLastEvent(),
							controlFlowAlignment.getLastTransition());
					if (value != VirtualVariable.UNCHANGED) {
						DataStateHelper.storeValue(variableAssignments, determineVariableName(true, variableName),
								value, varType, mapping.getStringDiscretizer());
					}
				}
			}
		}
		return variableAssignments;
	}

	private Number extractVariableFromResult(ILPProblem ilpProblem, Result result, final ScopedVariable scopedVar) {
		Number processValue = null;
		String ilpVariable = scopedVar.variable.getInternalName(scopedVar.scope);
		if (result != null) {
			processValue = result.get(ilpVariable);
		}
		if (processValue == null) {
			if (processValue == null && scopedVar.variable.getType() == String.class) {
				// There is always one value for a String variable 
				processValue = mapping.getStringDiscretizer().convertToInt(StringDiscretizer.ANY_VALUE);
			} else { // Take the lower bound for all other variables
				processValue = ilpProblem.getVarLowerBound(ilpVariable);
			}
			if (processValue == null) { // Take the upper bound, if lower bound unknown
				processValue = ilpProblem.getVarUpperBound(ilpVariable);
			}
			if (processValue == null) { // Fall back to zero otherwise
				processValue = 0;
			}
		}
		return processValue;
	}

	private static String determineVariableName(boolean isLastStep, String variableName) {
		if (isLastStep) {
			// Prime 
			return variableName.concat("'");
		} else {
			// Previous write operations
			return variableName;
		}
	}

	private AlignedTraces createFullAlignedTrace(TraceMapping traceMapping, ILPProblem ilpProblem, Result result,
			ImmutableList<Step> stepList, List<ScopedVariable[]> variablesByStep) {

		final GenericTrace logTrace = new GenericTrace(stepList.size(), traceMapping.getTraceName());
		final GenericTrace processTrace = new GenericTrace(stepList.size(), traceMapping.getTraceName());

		// ILP Problem was built backwards
		variablesByStep = Lists.reverse(variablesByStep);

		int stepNr = 0;
		for (Step step : stepList) {

			final String activityName = step.getType() == StepTypes.L
					? traceMapping.getLogMapping().getEventClassifier().getClassIdentity(step.getLogNode())
					// workaround for synchronized name lookup in Transition.getLabel()
					: traceMapping.getLogMapping().getTransitionToName().get(step.getModelNode());

			ExecutionStep logStep = ExecutionStep.bottomStep;
			ExecutionStep processStep = ExecutionStep.bottomStep;
			switch (step.getType()) {
				case L :
					logStep = new ExecutionStep(activityName, step.getLogNode());
					break;
				case LMGOOD :
				case LMNOGOOD :
					logStep = new ExecutionStep(activityName, step.getLogNode());
					for (String variableName : mapping.getVariablesToWrite().get(step.getModelNode())) {
						Variable mappingVariable = mapping.getVariables().get(variableName);
						if (!mappingVariable.isVirtual() || includeVirtualVariablesInTraces) {
							Object value = traceMapping.getVariableAssignmentFor(variableName, step.getRecordForStep(),
									step.getLogNode(), step.getModelNode());
							if (value != null) {
								logStep.put(variableName, value);
							}
						}
					}
					//$FALL-THROUGH$ // log step already initialized!
				case MREAL :
				case MINVI :
					processStep = new ExecutionStep(activityName, step.getModelNode());

					// Handle virtual and final variables
					for (String variable : mapping.getVariablesToWrite().get(step.getModelNode())) {
						Variable mappingVariable = mapping.getVariables().get(variable);

						if (mappingVariable.isVirtual()) {
							if (includeVirtualVariablesInTraces) {
								Object value = traceMapping.getVariableAssignmentFor(mappingVariable.getName(),
										step.getRecordForStep(), step.getLogNode(), step.getModelNode());
								if (value != null) {
									// Virtual variable can be computed
									if (logStep == ExecutionStep.bottomStep) {
										// "Move in Data"
										logStep = new ExecutionStep();
										logStep.put(mappingVariable.getName(), value);
									}
									if (mapping.getVariableCost().isFinal(activityName, mappingVariable)) {
										// Also put as process value, in case this variable is final
										processStep.put(mappingVariable.getName(), value);
									}
								} else {
									// Not result, i.e. missing virtual variable currently treated as OK
								}
							}
						} else if (mapping.getVariableCost().isFinal(activityName, mappingVariable)) {
							// Only a final variable, just put the log value that was added earlier
							Object logValue = logStep.get(mappingVariable.getName());
							assert logValue != null : "Final variable " + mappingVariable.getName()
									+ " is NULL while creating final result";
							processStep.put(mappingVariable.getName(), logValue);
						}
					}

					if (step.getType() == StepTypes.MINVI) {
						processStep.setInvisible(true);
					}
					ScopedVariable[] mappingVariables = variablesByStep.get(stepNr++);
					for (int j = 0; j < mappingVariables.length; j++) {
						ScopedVariable scopedVar = mappingVariables[j];
						String processVariable = scopedVar.variable.getName();
						Class<?> varType = scopedVar.variable.getType();
						Number processValue = null;
						String internalName = scopedVar.variable.getInternalName(scopedVar.scope);
						if (result != null) {
							processValue = result.get(internalName);
						}
						if (processValue == null) {
							if (processValue == null && scopedVar.variable.getType() == String.class) {
								// There is always one value for a String variable 
								processValue = mapping.getStringDiscretizer().convertToInt(StringDiscretizer.ANY_VALUE);
							} else { // Take the lower bound for all other variables
								processValue = ilpProblem.getVarLowerBound(internalName);
							}
							if (processValue == null) { // Take the upper bound, if lower bound unknown
								processValue = ilpProblem.getVarUpperBound(internalName);
							}
							if (processValue == null) { // Fall back to zero otherwise
								processValue = 0;
							}
						}
						Variable variable = mapping.getVariables().get(processVariable);
						if (!variable.isVirtual() || includeVirtualVariablesInTraces) {
							if (logStep.containsKey(processVariable)) {
								Object logObjValue = logStep.get(processVariable);
								double logValue = DataStateHelper.convertToIlpValue(processVariable, logObjValue,
										mapping.getStringDiscretizer());
								if (DoubleMath.fuzzyEquals(logValue, processValue.doubleValue(), SOLVING_PRECISION))
									processValue = logValue;
							}
							processStep.put(processVariable, DataStateHelper.convertFromIlpValue(processVariable,
									processValue.doubleValue(), varType, mapping.getStringDiscretizer()));
						}
					}
					break;
				default :
					throw new IllegalArgumentException("Unkown step type " + step.getType());
			}
			logTrace.add(logStep);
			processTrace.add(processStep);
		}
		return new AlignedTraces() {

			public GenericTrace getLogTrace() {
				return logTrace;
			}

			public GenericTrace getProcessTrace() {
				return processTrace;
			}
		};
	}

	private double addAllVariableAssignments(TraceMapping traceMapping, DRecord record, ILPProblem problem, Step step,
			ScopedVariable[] mappingVariables, VariableAwareWriteOperationCost writeOperationCost) {
		boolean invisibleStep = (step.getType() == StepTypes.MINVI);
		double additionalCost = 0;
		for (int i = 0; i < mappingVariables.length; i++) {
			ScopedVariable scopedVar = mappingVariables[i];
			MILPVariable var = scopedVar.variable;
			Object attrValue = traceMapping.getVariableAssignmentFor(var.getName(), record, step.getLogNode(),
					step.getModelNode());
			String transitionLabel = step.getModelNode().getLabel(); //TODO this should not be the label!! but the ID
			if (attrValue != null) {
				// We have seen a value
				double variableValue = DataStateHelper.convertToIlpValue(var.getName(), attrValue,
						mapping.getStringDiscretizer());
				String ilpVarName = scopedVar.getIlpName();
				String boolVarName = var.getBooleanVarName(scopedVar.scope);
				addVariableAssignment(problem, ilpVarName, boolVarName, variableValue);
				problem.addObjective(writeOperationCost.costFaultyValue(transitionLabel, var.getVariable()),
						boolVarName);

				if (problem.hasNullConstraint(ilpVarName)) {
					// Some guard contained a comparison to NULL for this variable. Record the fact that we have seen a value.  
					additionalCost += problem.writeNullConstraint(ilpVarName, false);
				}

			} else if (!invisibleStep) {
				// Missing write 
				additionalCost += writeOperationCost.notWritingCost(transitionLabel, var.getVariable());
			}
		}

		return additionalCost;
	}

	private static void addVariableAssignment(ILPProblem problem, String ilpVarName, String boolVar,
			double variableValue) {
		problem.setVarType(boolVar, Boolean.class);
		double upperbound = calcUpperBound(problem, ilpVarName, variableValue);
		Linear constraint = new Linear();
		constraint.add(1, ilpVarName);
		constraint.add(-1.0 * upperbound, boolVar);
		problem.addAssignmentConstraint(constraint, Operator.LE, variableValue);
		constraint = new Linear();
		constraint.add(-1, ilpVarName);
		constraint.add(-1.0 * upperbound, boolVar);
		problem.addAssignmentConstraint(constraint, Operator.LE, -1.0 * variableValue);
	}

	private static double calcUpperBound(ILPProblem problem, String ilpVarName, double value) {
		double upperbound;
		Class<?> varType = problem.getVarType(ilpVarName);
		if (varType == Boolean.class) {
			upperbound = 1;
		} else {
			Number varUpperBound = problem.getVarUpperBound(ilpVarName);
			if (varUpperBound != null) {
				upperbound = varUpperBound.doubleValue();
			} else {
				upperbound = value;
			}
		}
		return upperbound;
	}

	// ******************************* Assertion Code *****************************

	private boolean reCheckILP(ILPProblem ilpProblem, Result result, Solver solver) {
		Result result2 = ilpProblem.solve(solver);
		if ((result != null && result2 != null)
				&& ((result == null && result2 != null) || (result != null && result2 == null)
						|| !Objects.equals(result.getObjective(), result2.getObjective()))) {
			System.out.println("Second round of ILP solving does not agree with first one!");
			System.out.println("Problem: " + ilpProblem);
			System.out.println("1st round: " + result);
			System.out.println("2nd round: " + result2);
			return false;
		} else {
			return true;
		}
	}

	private boolean lastGuardTrue(TraceMapping mapping2, ControlFlowAlignmentResult controlFlowAlignment,
			List<ScopedVariable[]> mappingVariablesToConditionStep, ILPProblem ilpProblem, Result result,
			boolean treatMissingVariablesAsFalse) {

		DataState variableAssignments = createPartialAlignedTrace(mapping2, ilpProblem, result, 0, controlFlowAlignment,
				ImmutableList.<Step>copyOf(controlFlowAlignment.getSteps()), mappingVariablesToConditionStep)
						.getVariableAssignment();

		if (controlFlowAlignment.getLastTransition() != null) {
			GuardExpression guard = mapping.getGuardExpression(controlFlowAlignment.getLastTransition());
			if (guard != null) {
				if (ExpressionEvaluator.checkAllVariableNotNull(guard, variableAssignments)) {
					boolean guardOk = ExpressionEvaluator.isTrue(guard, variableAssignments,
							LogMapping.convertToVariableTypes(mapping.getVariables()), mapping.getStringDiscretizer());
					if (!guardOk) {
						return false;
					}
				} else {
					if (treatMissingVariablesAsFalse) {
						return false;
					}
				}
			}
		}

		return true;
	}

	private boolean areAllGuardsTrue(TraceMapping traceMapping, ControlFlowAlignmentResult controlFlowAlignment,
			List<ScopedVariable[]> mappingVariablesToConditionStep, ILPProblem ilpProblem, Result result,
			boolean treatMissingVariablesAsFalse) {
		DataState variableAssignments = dataStateFactory.newDataState();
		int modelStepNr = 0;

		List<ScopedVariable[]> reversedVariables = Lists.reverse(mappingVariablesToConditionStep);

		for (Iterator<ControlFlowAlignmentResult.Step> iterator = controlFlowAlignment.getSteps().iterator(); iterator
				.hasNext();) {
			Step step = iterator.next();
			switch (step.getType()) {
				case L :
					break;
				case LMGOOD :
				case LMNOGOOD :
				case MINVI :
				case MREAL :

					ScopedVariable[] mappingVariables = reversedVariables.get(modelStepNr++);
					for (int j = 0; j < mappingVariables.length; j++) {
						ScopedVariable scopedVar = mappingVariables[j];
						Number processValue = extractVariableFromResult(ilpProblem, result, scopedVar);
						String processVariable = scopedVar.variable.getName();
						String varName = determineVariableName(true, processVariable);
						Class<?> varType = scopedVar.variable.getType();
						DataStateHelper.storeValueFromIlp(variableAssignments, varName, processValue, varType);
					}

					GuardExpression guard = mapping.getGuardExpression(step.getModelNode());
					if (guard != null) {
						if (ExpressionEvaluator.checkAllVariableNotNull(guard, variableAssignments)) {
							boolean guardOk = ExpressionEvaluator.isTrue(guard, variableAssignments,
									LogMapping.convertToVariableTypes(mapping.getVariables()),
									mapping.getStringDiscretizer());
							if (!guardOk) {
								return false;
							}
						} else {
							if (treatMissingVariablesAsFalse) {
								return false;
							}
						}
					}

					break;
				default :
					throw new IllegalArgumentException("Unknown step " + step.getType());
			}
			variableAssignments = variableAssignments.clone(true);
		}
		return true;
	}

	private String assertErrorGuardsViolatedAssertionMessage(TraceMapping traceMapping,
			ControlFlowAlignmentResult controlFlowAlignment, ILPProblem ilpProblem, Result result) {
		return String.format(
				"ILP Solver returned invalid result for %s, a guard is NOT satisfied, but ILP problem is feasible! \nILP Problem: %s \nILP Solution: %s",
				controlFlowAlignment, ilpProblem, result);
	}

	private String assertErrorGuardsNotViolatedAssertionMessage(TraceMapping traceMapping,
			ControlFlowAlignmentResult controlFlowAlignment, ILPProblem ilpProblem, Result result) {
		return String.format(
				"ILP Solver returned invalid result for %s, a guard is satisfied, but ILP problem is NOT feasible! \nILP Problem: %s \nILP Solution: %s",
				controlFlowAlignment, ilpProblem, result);
	}

	private String assertErrorFullDataAlignmentDisagrees(TraceMapping mapping,
			ControlFlowAlignmentResult controlFlowAlignment, ILPProblem ilpProblem, Result result, int ilpCost,
			PartialDataAlignedTrace partialDataAlignedTrace) {
		try {
			FullDataAlignedTrace fullAlignment = createAlignment(mapping, controlFlowAlignment);
			System.err.println("Full alignment: " + fullAlignment);
			System.err.println("using ILP: " + createFullILPProblem(mapping, controlFlowAlignment));
			System.err.println("Partial alignment: " + partialDataAlignedTrace);
			System.err.println("with discretizer: " + mapping.getLogMapping().getStringDiscretizer().toString());
			return String.format(
					"Partial ILP problem returned invalid result for %s,\n full ILP problem returns data cost %s instead of data cost %s! \nILP Problem: %s \nILP Solution: %s",
					controlFlowAlignment, fullAlignment.getDataCost(), ilpCost, ilpProblem, result);
		} catch (MILPProblemException e) {
			return Throwables.getStackTraceAsString(e);
		}
	}

	private ILPProblem createFullILPProblem(TraceMapping mapping, ControlFlowAlignmentResult controlFlowAlignment) {
		List<ScopedVariable[]> variablesByStep = new ArrayList<>(controlFlowAlignment.getApproximateSize());
		ImmutableList<Step> stepList = ImmutableList.copyOf(controlFlowAlignment.getSteps());
		TObjectIntMap<String> timesWrittenVariable = new TObjectIntHashMap<>(variables.size(), 0.25f, 0);

		ILPProblem ilpProblem = new ILPProblemImpl();
		addGuardsAndAssignments(mapping, stepList, controlFlowAlignment.getLastRecord(), timesWrittenVariable,
				ilpProblem, variablesByStep, CalculationMode.FULL);
		addDefaultOrFreeVariablesToILP(timesWrittenVariable, ilpProblem);

		Solver solver = getSolver();
		ilpProblem.solve(solver);

		return ilpProblem;
	}

	// ******************************* Assertion Code *****************************

}
