package org.processmining.plugins.balancedconformance.dataflow.datastate;

import java.util.Arrays;

final class DataStateImpl extends AbstractDataStateImpl {

	interface VariableIndexer {

		int getIndex(String varName);

		int getNormalIndex(int primeIndex);

		int getPrimeIndex(int normalIndex);

		int getFirstPrimeIndex();

		int sizeWithPrimes();

		int sizeNormal();

		String getVarName(int index);

		int sizeBitset();

		int sizeIncludingBitset();

	}

	private final VariableIndexer mapping;
	private final long[] values;
	private int hash = 0;

	public DataStateImpl(final VariableIndexer mapping) {
		super();
		this.mapping = mapping;
		this.values = new long[mapping.sizeIncludingBitset()];
	}

	DataStateImpl(final VariableIndexer mapping, long[] values) {
		super();
		this.mapping = mapping;
		this.values = values;
	}

	// ***********************************
	// Borrowed from Lucene project (OpenBitSet) under Apache 2.0 license 
	private final void mark(int index) {
		assert index < mapping.sizeWithPrimes();
		int wordNum = index >> 6;
		long bitmask = 1L << index;
		assert wordNum < mapping.sizeBitset();
		values[wordNum] |= bitmask;
	}

	private final void unmark(int index) {
		assert index < mapping.sizeWithPrimes();
		int wordNum = index >> 6;
		long bitmask = 1L << index;
		assert wordNum < mapping.sizeBitset();
		values[wordNum] &= ~bitmask;
	}

	private final boolean isMarked(int index) {
		assert index < mapping.sizeWithPrimes();
		int wordNum = index >> 6;
		long bitmask = 1L << index;
		assert wordNum < mapping.sizeBitset();
		return (values[wordNum] & bitmask) != 0;
	}

	private final int nextSetBit(int index, int sizeBitset) {
		if (values.length == 0) {
			return -1;
		}
		int wordNum = index >> 6;
		long word = values[wordNum] >> index; // skip all the bits to the right of index
		if (word != 0) {
			return index + Long.numberOfTrailingZeros(word);
		}
		while (++wordNum < sizeBitset) {
			word = values[wordNum];
			if (word != 0) {
				return (wordNum << 6) + Long.numberOfTrailingZeros(word);
			}
		}
		return -1;
	}

	// ***********************************

	@Override
	public boolean isEmpty() {
		return size() == 0;
	}

	@Override
	public int size() {
		// Borrowed from Lucene project (BitUtil) under Apache 2.0 license 
		final int sizeBitset = mapping.sizeBitset();
		int count = 0;
		for (int i = 0, end = sizeBitset; i < end; ++i) {
			count += Long.bitCount(values[i]);
		}
		return count;
	}

	@Override
	public DataState clone(final boolean movePrimes) {
		if (movePrimes) {
			final DataStateImpl clone = new DataStateImpl(mapping, Arrays.copyOf(values, values.length));
			int sizeBitset = mapping.sizeBitset();
			for (int i = nextSetBit(mapping.getFirstPrimeIndex(), sizeBitset); i >= 0; i = nextSetBit(i + 1, sizeBitset)) {
				int correctedIndex = i + sizeBitset;
				// Get prime value
				long value = values[correctedIndex];
				// Remove prime value
				clone.unmark(i);
				clone.values[correctedIndex] = 0; // better compression
				// Set non-prime value
				int nonPrimeIndex = mapping.getNormalIndex(i);
				clone.values[nonPrimeIndex + sizeBitset] = value;
				clone.mark(nonPrimeIndex);
			}
			return clone;
		} else {
			return new DataStateImpl(mapping, Arrays.copyOf(values, values.length));
		}
	}

	@Override
	public void movePrimes() {
		int sizeBitset = mapping.sizeBitset();
		for (int i = nextSetBit(mapping.getFirstPrimeIndex(), sizeBitset); i >= 0; i = nextSetBit(i + 1, sizeBitset)) {
			final int normalIndex = mapping.getNormalIndex(i);
			int correctedIndex = i  + sizeBitset;
			values[normalIndex + sizeBitset] = values[correctedIndex];
			values[correctedIndex] = 0; // better compression
			unmark(i);
		}
	}

	@Override
	public DataState makeStateImmutable() {
		hash = hashCode();
		return this;
	}

	@Override
	protected void putInternal(final String varName, final long value) {
		final Integer index = mapping.getIndex(varName); 
		mark(index); //mark original index 
		values[index + mapping.sizeBitset()] = value; // adjust for bitset
	}

	@Override
	public final boolean contains(String varName) {
		return isMarked(mapping.getIndex(varName));
	}

	@Override
	protected Long tryGetInternal(final String varName) {
		final int index = mapping.getIndex(varName);
		if (isMarked(index)) {
			return values[index + mapping.sizeBitset()];  // adjust for bitset
		} else {
			return null;
		}
	}

	protected Long tryGetPrimeOrNormalInternal(final String varName) {
		int normalIndex = mapping.getIndex(varName);
		final int primeIndex = mapping.getPrimeIndex(normalIndex); 	
		if (isMarked(primeIndex)) {
			return values[primeIndex + mapping.sizeBitset()];  // adjust for bitset
		} else {
			if (isMarked(normalIndex)) {
				return values[normalIndex + mapping.sizeBitset()];   // adjust for bitset
			} else {
				return null;
			}
		}
	}

	@Override
	public int hashCode() {
		if (hash != 0) {
			return hash;
		}
		return Arrays.hashCode(values);
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final DataStateImpl other = (DataStateImpl) obj;
		if (!valuesEquals(other)) {
			return false;
		}
		return true;
	}

	private final boolean valuesEquals(final DataStateImpl other) {
		for (int i = 0; i < values.length; i++) {
			if (values[i] != other.values[i]) {
				return false;
			}
		}
		return true;
	}

	@Override
	public String toString() {
		final StringBuilder sb = new StringBuilder("[");
		int sizeBitset = mapping.sizeBitset();
		for (int i = nextSetBit(0, sizeBitset); i >= 0; i = nextSetBit(i + 1, sizeBitset)) {
			long value = values[i + sizeBitset];
			final String varName = mapping.getVarName(i);
			sb.append(varName);
			sb.append("=");
			sb.append(String.valueOf(value));
			sb.append(", ");
		}
		if (sb.length() > 1) {
			sb.delete(sb.length() - 2, sb.length());
		}
		sb.append("]");
		return sb.toString();
	}

	public Object getData() {
		return values;
	}

}