/*
* Adapted from SolverGurobi.java of Java ILP
*
* Java ILP is free software: you can redistribute it and/or modify it under
* the terms of the GNU Lesser General Public License as published by the
* Free Software Foundation, either version 3 of the License, or (at your
* option) any later version.
*
* Java ILP is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
* License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with Java ILP. If not, see http://www.gnu.org/licenses/.
*/ 
package org.processmining.plugins.balancedconformance.dataflow.lp.gurobi;

import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.google.common.collect.Maps;

import gurobi.GRB;
import gurobi.GRBEnv;
import gurobi.GRBException;
import gurobi.GRBLinExpr;
import gurobi.GRBModel;
import gurobi.GRBVar;
import net.sf.javailp.AbstractSolver;
import net.sf.javailp.Constraint;
import net.sf.javailp.Linear;
import net.sf.javailp.Operator;
import net.sf.javailp.OptType;
import net.sf.javailp.Problem;
import net.sf.javailp.Result;
import net.sf.javailp.ResultImpl;
import net.sf.javailp.Term;
import net.sf.javailp.VarType;

/**
 * The {@code MyGurobiSolver} is the {@code Solver} Gurobi.
 * 
 * @author fabiogenoese, lukasiewycz, F. Mannhardt
 *
 */
public class DataAlignmentGurobiSolver extends AbstractSolver {
	
	private GRBEnv env;

	/**
	 * @param env gurobi environment
	 */
	public DataAlignmentGurobiSolver(GRBEnv env) {
		super();
		this.env = env;
	}

	/* (non-Javadoc)
	 * @see net.sf.javailp.Solver#solve(net.sf.javailp.Problem)
	 */
	public Result solve(Problem problem) {
		Map<Object, GRBVar> objToVar = Maps.newHashMapWithExpectedSize(problem.getVariables().size());
		Map<GRBVar, Object> varToObj = Maps.newHashMapWithExpectedSize(problem.getVariables().size());
		try {
			GRBModel model = new GRBModel(env);
			OptType optType = problem.getOptType();
			Map<Object, Double> optimizationCoefficients = Maps.newHashMapWithExpectedSize(problem.getObjective().size());
			Linear objective = problem.getObjective();
			if (objective != null) {
				for (Term term : objective) {
					Object variable = term.getVariable();
					double coeff = term.getCoefficient().doubleValue();
					if (optType == OptType.MAX) {
						coeff *= -1;
					}
					optimizationCoefficients.put(variable, coeff);
				}
			}
			for (Object variable : problem.getVariables()) {
				VarType varType = problem.getVarType(variable);
				Number lowerBound = problem.getVarLowerBound(variable);
				Number upperBound = problem.getVarUpperBound(variable);
				double lb = (lowerBound != null ? lowerBound.doubleValue() : -Double.MAX_VALUE);
				double ub = (upperBound != null ? upperBound.doubleValue() : Double.MAX_VALUE);
				final String name = variable.toString();
				final char type;
				switch (varType) {
					case BOOL :
						type = GRB.BINARY;
						break;
					case INT :
						type = GRB.INTEGER;
						break;
					default : // REAL
						type = GRB.CONTINUOUS;
						break;
				}
				Double coeff = optimizationCoefficients.get(variable);
				if (coeff == null) {
					coeff = 0.0;
				}
				GRBVar var = model.addVar(lb, ub, coeff, type, name);
				objToVar.put(variable, var);
				varToObj.put(var, variable);
			}
			model.update();
			int currentName = 1;
			for (Constraint constraint : problem.getConstraints()) {
				GRBLinExpr expr = new GRBLinExpr();
				for (Term term : constraint.getLhs()) {
					GRBVar var = objToVar.get(term.getVariable());
					expr.addTerm(term.getCoefficient().doubleValue(), var);
				}
				final char operator;
				if (constraint.getOperator() == Operator.GE)
					operator = GRB.GREATER_EQUAL;
				else if (constraint.getOperator() == Operator.LE)
					operator = GRB.LESS_EQUAL;
				else
					operator = GRB.EQUAL;				
				model.addConstr(expr, operator, constraint.getRhs().doubleValue(), "c".concat(String.valueOf(currentName++)));
			}
			for (Hook hook : hooks) {
				hook.call(env, model, objToVar, varToObj, problem);
			}
			model.optimize();
			if (model.get(GRB.IntAttr.Status) == GRB.OPTIMAL) {
				Result result;
				if (problem.getObjective() != null) {
					result = new ResultImpl(problem.getObjective());
				} else {
					result = new ResultImpl();
				}
				for (Entry<Object, GRBVar> entry : objToVar.entrySet()) {
					Object variable = entry.getKey();
					GRBVar var = entry.getValue();
					try {
						double primalValue = var.get(GRB.DoubleAttr.X);
						if (problem.getVarType(variable).isInt()) {
							int v = (int) Math.round(primalValue);
							result.putPrimalValue(variable, v);
						} else {
							result.putPrimalValue(variable, primalValue);
						}
					} catch (GRBException e) {
						System.out.println("Variable "+variable+", Error: "+e.getErrorCode()+": "+e.getMessage());
						System.out.println("Problem: "+ problem.toString());
						model.write("error"+String.valueOf(System.currentTimeMillis())+".mps");
						return null;
					}									
				}
				return result;
			} else {
				return null;
			}
		} catch (GRBException e) {
			System.out.println(e.getErrorCode()+": "+e.getMessage());
			System.out.println("Problem: "+ problem.toString());
			return null;
		}
	}

	/**
	 * The {@code Hook} for the {@code SolverGurobi}.
	 * 
	 * @author lukasiewycz
	 * 
	 */
	public interface Hook {
		/**
		 * This method is called once before the optimization and allows to
		 * change some internal settings.
		 * 
		 * @param env
		 *            the environment
		 * @param model
		 *            the model
		 * @param objToVar
		 *            the map from objects to gurobi variables
		 * @param varToObj
		 *            the map from gurobi variables to objects
		 * @param problem
		 *            the problem
		 */
		public void call(GRBEnv env, GRBModel model, Map<Object, GRBVar> objToVar, Map<GRBVar, Object> varToObj,
				Problem problem);
	}

	protected final Set<Hook> hooks = new HashSet<>();

	/**
	 * Adds a hook.
	 * 
	 * @param hook
	 *            the hook to be added
	 */
	public void addHook(Hook hook) {
		hooks.add(hook);
	}

	/**
	 * Removes a hook
	 * 
	 * @param hook
	 *            the hook to be removed
	 */
	public void removeHook(Hook hook) {
		hooks.remove(hook);
	}

}
