package org.processmining.plugins.balancedconformance.dataflow.lp.lpsolve;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import org.processmining.plugins.balancedconformance.dataflow.MILPProblemBuilder;

import lpsolve.LpSolve;
import lpsolve.LpSolveException;
import net.sf.javailp.AbstractSolverFactory;
import net.sf.javailp.Solver;

public final class DataAlignmentSolverFactoryLpSolve extends AbstractSolverFactory implements AutoCloseable {

	static final class LpSolveInstance {
		LpSolve lp;
		int timesUsed;
		public LpSolveInstance(LpSolve lp) {
			this.lp = lp;
		}
	}

	private static final int VERBOSE = 0;
	private static final int MAX_USAGE = 128;

	static {
		System.loadLibrary("lpsolve55");
		System.loadLibrary("lpsolve55j");
	}

	private final BlockingQueue<LpSolveInstance> solverQueue;
	private DataAlignmentLpSolveSolver solver;

	public DataAlignmentSolverFactoryLpSolve(int concurrentThreads) {
		super();
		solverQueue = new LinkedBlockingQueue<>(concurrentThreads);
		try {
			for (int i = 0; i < concurrentThreads; i++) {
				solverQueue.add(new LpSolveInstance(createLp()));
			}
		} catch (LpSolveException e) {
			throw new RuntimeException("Cannot initialize LpSolve!", e);
		}
		solver = new DataAlignmentLpSolveSolver(this);
	}

	private LpSolve createLp() throws LpSolveException {
		LpSolve lp = LpSolve.makeLp(0, 0);
		lp.setEpsint(MILPProblemBuilder.SOLVING_PRECISION);
		lp.setTimeout(MILPProblemBuilder.TIMEOUT);
		lp.setVerbose(VERBOSE);
		lp.setScaling(LpSolve.SCALE_GEOMETRIC | LpSolve.SCALE_EQUILIBRATE | LpSolve.SCALE_INTEGERS);
		lp.setScalelimit(5);
		lp.setPivoting(LpSolve.PRICER_DEVEX | LpSolve.PRICE_ADAPTIVE);
		lp.setMaxpivot(250);
		lp.setBbFloorfirst(LpSolve.BRANCH_AUTOMATIC);
		lp.setBbRule(LpSolve.NODE_PSEUDONONINTSELECT | LpSolve.NODE_GREEDYMODE | LpSolve.NODE_DYNAMICMODE
				| LpSolve.NODE_RCOSTFIXING);  
		lp.setBbDepthlimit(-50);
		lp.setObjBound(1e30);
		lp.setBreakAtValue(-1e30);
		lp.setAntiDegen(LpSolve.ANTIDEGEN_FIXEDVARS | LpSolve.ANTIDEGEN_STALLING);
		lp.setImprove(LpSolve.IMPROVE_DUALFEAS | LpSolve.IMPROVE_THETAGAP);
		lp.setBasiscrash(LpSolve.CRASH_NOTHING);
		lp.setSimplextype(LpSolve.SIMPLEX_DUAL_PRIMAL);
		// LpSolve.PRESOLVE_COLS messes up variable indexing
		// LpSolve.PRESOLVE_ROWS messes up problems without objective function and some others
		return lp;
	}

	@Override
	public Solver get() {
		return getInternal();
	}

	@Override
	protected Solver getInternal() {
		return solver;
	}

	protected void finalize() throws Throwable {
		close();
		super.finalize();
	}

	@Override
	public void close() {
		while (!solverQueue.isEmpty()) {
			solverQueue.poll().lp.deleteAndRemoveLp();
		}
	}

	public LpSolveInstance getLp() {
		try {
			LpSolveInstance instance = solverQueue.poll();
			if (instance.timesUsed == MAX_USAGE) {
				instance.lp.deleteAndRemoveLp();
				instance.lp = createLp();
				instance.timesUsed = 0;
			} else if (instance.timesUsed != 0) {				
				// prepare
				instance.lp.unscale();
				instance.lp.defaultBasis();
				instance.lp.resetBasis();
				instance.lp.resizeLp(0, 0);
			}
			instance.timesUsed++;
			return instance;
		} catch (LpSolveException e) {
			e.printStackTrace();
			throw new RuntimeException(e);
		}
	}

	@SuppressWarnings("unused")
	public void returnLp(LpSolveInstance lp) throws InterruptedException {
		solverQueue.add(lp);
	}

}