package org.processmining.plugins.balancedconformance.filter;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.DataConformance.ResultReplay;

public class FilterLogByReplayResultPlugin {

	private static interface TraceFilter {
		boolean includeTrace(String traceName, ResultReplay replayResult);
	}

	@Plugin(name = "Filter Log: Retain Trace with Higher Fitness", level = PluginLevel.PeerReviewed, parameterLabels = {
			"Log", "ResultReplay" }, returnLabels = { "Log" }, returnTypes = {
					XLog.class }, userAccessible = true, categories = {
							PluginCategory.Filtering }, help = "Filter an event log to retain traces with a high replay fitness.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl", pack = "DataAwareReplayer")
	public XLog filterLogByFitnessHigher(final UIPluginContext context, XLog log, ResultReplay replayResult)
			throws UserCancelledException {

		try {
			final double fitnessLimit = ProMUIHelper.queryForDouble(context, "Specify Minimal Fitness Level (inclusive)", 0.0, 1.0, 0.0);
			return filterLog(log, replayResult, new TraceFilter() {

				public boolean includeTrace(String traceName, ResultReplay replayResult) {
					Alignment a = replayResult.getAlignmentByTraceName(traceName);
					if (a != null) {
						return Double.compare(a.getFitness(), fitnessLimit) >= 0;
					} else {
						String errorMessage = "Could not find the corresponding alignment for trace with name "
								+ traceName;
						ProMUIHelper.showErrorMessage(context, errorMessage, "Error filtering log");
						throw new RuntimeException(errorMessage);
					}
				}
			}, context);
		} catch (NumberFormatException e) {
			String errorMessage = "Could not filter log, please specify fitness as a number with '.' as decimal separator ('0.5)' !";
			ProMUIHelper.showErrorMessage(context, errorMessage, "Error filtering log");
			throw new RuntimeException(errorMessage);
		}
	}

	@Plugin(name = "Filter Log: Retain Trace with Lower Fitness", level = PluginLevel.PeerReviewed, parameterLabels = {
			"Log", "ResultReplay" }, returnLabels = { "Log" }, returnTypes = {
					XLog.class }, userAccessible = true, categories = {
							PluginCategory.Filtering }, help = "Filter an event log to retain traces with a low replay fitness.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl", pack = "DataAwareReplayer")
	public XLog filterLogByFitnessLower(final UIPluginContext context, XLog log, ResultReplay replayResult)
			throws UserCancelledException {
		try {
			final double fitnessLimit = ProMUIHelper.queryForDoubleZeroOne(context, "Specify Maximal Fitness Level (inclusive)");
			return filterLog(log, replayResult, new TraceFilter() {

				public boolean includeTrace(String traceName, ResultReplay replayResult) {
					Alignment a = replayResult.getAlignmentByTraceName(traceName);
					if (a != null) {
						return Double.compare(a.getFitness(), fitnessLimit) <= 0;
					} else {
						String errorMessage = "Could not find the corresponding alignment for trace with name "
								+ traceName;
						ProMUIHelper.showErrorMessage(context, errorMessage, "Error filtering log");
						throw new RuntimeException(errorMessage);
					}
				}
			}, context);
		} catch (NumberFormatException e) {
			String errorMessage = "Could not filter log, please specify fitness as a number with '.' as decimal separator ('0.5)' !";
			ProMUIHelper.showErrorMessage(context, errorMessage, "Error filtering log");
			throw new RuntimeException(errorMessage);
		}
	}

	private XLog filterLog(XLog log, ResultReplay replayResult, TraceFilter traceFilter, UIPluginContext context) {
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog filteredLog = factory.createLog((XAttributeMap) log.getAttributes().clone());

		for (XTrace xTrace : log) {
			String traceName = XConceptExtension.instance().extractName(xTrace);
			if (traceName == null) {
				String errorMessage = "Trace " + XUtils.stringifyTrace(xTrace)
						+ " is missing a concept:name! Cannot filter a log without unique trace names.";
				ProMUIHelper.showErrorMessage(context, errorMessage, "Error filtering log");
				throw new RuntimeException(errorMessage);
			} else if (traceFilter.includeTrace(traceName, replayResult)) {
				filteredLog.add(xTrace);
			}
		}
		return filteredLog;
	}

}
