/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.mapping;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.log.utils.XUtils;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.plugins.DataConformance.framework.WriteOperationCost;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration.UnassignedMode;
import org.processmining.plugins.balancedconformance.dataflow.IlpExpressionTemplate;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateHelper;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.functions.FunctionsTransformations;
import org.processmining.plugins.balancedconformance.functions.VirtualVariable;
import org.processmining.plugins.balancedconformance.functions.VirtualVariableAwareWriteOperationCostWrapper;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;

import com.google.common.base.Function;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableMap.Builder;
import com.google.common.collect.ImmutableSetMultimap;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.Maps;
import com.google.common.collect.Multimaps;
import com.google.common.collect.SetMultimap;
import com.google.common.collect.Sets;

/**
 * All information about the {@link XLog} that is used during the Data
 * Alignment.
 * 
 * @author F. Mannhardt
 * 
 */
public final class LogMapping {

	private final Map<String, Variable> variables = new HashMap<>();
	private final SetMultimap<String, Variable> reverseVariableMapping;
	private final VariableAwareWriteOperationCost variableCost;

	private final UnassignedMode unassignedMode;

	private final Map<Transition, String> transitionToName;
	private final SetMultimap<String, Transition> nameToTransition;
	private final SetMultimap<String, Transition> eventIdentityToTransition;

	private final Map<Transition, Boolean> hasWriteOps;
	private final Map<Transition, LinkedHashSet<String>> variablesToWrite;
	private final Map<Transition, GuardExpression> guardExpressions = new HashMap<>();
	private final Set<Transition> finalVariableInGuard;
	private final Map<Transition, IlpExpressionTemplate> transitionToExpression;

	private final TransEvClassMapping transEvClassMapping;
	private final XEventClassifier eventClassifier;
	private final XEventClasses eventClasses;
	private final Map<XEventClass, Integer> eventClass2Cost;
	private final Map<Transition, Integer> transition2Cost;

	private final StringDiscretizer stringDiscretizer;
	private boolean hasVirtualVariables = false;

	@Deprecated
	public LogMapping(TransEvClassMapping transEvClassMapping, Map<String, String> variableMapping,
			WriteOperationCost writeOperationCost, Map<String, Object> upperBounds, Map<String, Object> lowerBounds,
			Map<String, Object> defaultValues, UnassignedMode unassignedMode, DataPetriNet net, XLog log)
			throws DataAlignmentException {
		this(transEvClassMapping, null, null, variableMapping, writeOperationCost, defaultValues,
				defaultValues, defaultValues, unassignedMode, net, log);
	}

	public LogMapping(TransEvClassMapping transEvClassMapping, Map<XEventClass, Integer> eventClass2Cost,
			Map<Transition, Integer> transition2Cost, Map<String, String> variableMapping,
			WriteOperationCost writeOperationCost, Map<String, Object> upperBounds, Map<String, Object> lowerBounds,
			Map<String, Object> defaultValues, UnassignedMode unassignedMode, DataPetriNet net, XLog log)
			throws DataAlignmentException {
		super();
		this.eventClass2Cost = eventClass2Cost;
		this.transition2Cost = transition2Cost;
		this.unassignedMode = unassignedMode;
		this.eventClassifier = transEvClassMapping.getEventClassifier();
		this.transEvClassMapping = transEvClassMapping;
		this.eventClasses = XUtils.createEventClasses(transEvClassMapping.getEventClassifier(), log);

		// Mapping for Transitions
		this.transitionToName = createTransitionsToNameMap(net);
		this.nameToTransition = Multimaps.invertFrom(Multimaps.forMap(transitionToName),
				HashMultimap.<String, Transition>create(variableMapping.size(), 4));
		this.eventIdentityToTransition = createEventClassToTransitionsMap(transEvClassMapping);

		// Mapping for Variables
		this.variablesToWrite = createVariablesToWriteMap(net);
		this.hasWriteOps = new HashMap<>();

		@SuppressWarnings("rawtypes")
		Map<String, Class> varType = PetriNetWithData.getVariableTypes(net);
		try {
			this.stringDiscretizer = createDiscretizer(net, log, variableMapping, varType, defaultValues,
					eventIdentityToTransition, variablesToWrite, eventClassifier);
		} catch (StringDiscretizerException e1) {
			throw new DataAlignmentException(e1);
		}

		final Set<String> variableUsedInGuard = new HashSet<>();
		for (Transition t : net.getTransitions()) {
			if (t instanceof PNWDTransition) {
				PNWDTransition pnwdTransition = (PNWDTransition) t;
				hasWriteOps.put(t, pnwdTransition.getWriteOperations().size() > 0);
				if (pnwdTransition.getGuardExpression() != null) {
					GuardExpression guardExpression = pnwdTransition.getGuardExpression();
					variableUsedInGuard.addAll(guardExpression.getNormalVariables());
					variableUsedInGuard.addAll(pnwdTransition.getGuardExpression().getPrimeVariables());
				}
			} else {
				hasWriteOps.put(t, false);
			}
		}

		for (@SuppressWarnings("rawtypes")
		Entry<String, Class> entry : varType.entrySet()) {

			String variableName = entry.getKey();
			Class<?> type = entry.getValue();

			Variable variable = new Variable(variableName, type);
			variable.setAttributeName(variableMapping.get(variableName));
			variable.setLowerBound(lowerBounds.get(variableName));
			variable.setUpperBound(upperBounds.get(variableName));

			DataElement dataElement = net.getVariable(variableName);

			Object upperBound = upperBounds.containsKey(variableName) ? upperBounds.get(variableName)
					: dataElement.getMaxValue();
			if (upperBound != null || type == String.class || type == Date.class) {
				if (type == String.class) {
					// Special treatment for String type as there is a special upper and lower bound (user cannot specify this)
					variable.setIlpUpperBound(Double.valueOf(stringDiscretizer.getUpperBound()));
				} else if (type == Date.class) {
					// Will work until today :)
					variable.setIlpUpperBound(
							DataStateHelper.convertToIlpValue(variableName, new Date(), stringDiscretizer));
				} else {
					variable.setIlpUpperBound(
							DataStateHelper.convertToIlpValue(variableName, upperBound, stringDiscretizer));
				}
			}

			Object lowerBound = lowerBounds.containsKey(variableName) ? lowerBounds.get(variableName)
					: dataElement.getMinValue();
			if (lowerBound != null || type == String.class || type == Date.class) {
				if (type == String.class) {
					// Special treatment for String type as there is a special upper and lower bound (user cannot specify this)
					variable.setIlpLowerBound(Double.valueOf(stringDiscretizer.getLowerBound()));
				} else if (type == Date.class) {
					variable.setIlpLowerBound(DataStateHelper.convertToIlpValue(variableName, 0, stringDiscretizer));
				} else {
					variable.setIlpLowerBound(
							DataStateHelper.convertToIlpValue(variableName, lowerBound, stringDiscretizer));
				}
			}
			Object defaultValue = defaultValues.get(variableName);
			if (defaultValue != null) {
				variable.setDefaultValue(defaultValue);
				variable.setIlpDefaultValue(
						DataStateHelper.convertToIlpValue(variableName, defaultValue, stringDiscretizer));
			}
			variable.setUsedInGuard(variableUsedInGuard.contains(variableName));

			variables.put(variable.getName().intern(), variable);
		}

		// Multiple variables in our DPN-net may be mapped to the same attribute name in the event log 
		this.reverseVariableMapping = HashMultimap.create();
		for (Entry<String, String> entry : variableMapping.entrySet()) {
			reverseVariableMapping.put(entry.getValue(), variables.get(entry.getKey()));
		}

		// Mapping for virtual variables that are results of function calls 
		SetMultimap<Transition, VirtualVariable> virtualVariables = LinkedHashMultimap.create();
		for (Transition t : net.getTransitions()) {
			if (t instanceof PNWDTransition) {
				PNWDTransition pnwdTransition = (PNWDTransition) t;
				if (FunctionsTransformations.hasFunction(pnwdTransition)) {
					guardExpressions.put(t, FunctionsTransformations.transformExpression(pnwdTransition,
							nameToTransition, virtualVariables, ImmutableMap.copyOf(variables), writeOperationCost));
				} else if (pnwdTransition.getGuardExpression() != null) {
					guardExpressions.put(t, pnwdTransition.getGuardExpression());
				}
			}
		}

		for (Entry<Transition, Collection<VirtualVariable>> varEntry : virtualVariables.asMap().entrySet()) {

			hasVirtualVariables = true;

			Transition transition = varEntry.getKey();

			for (VirtualVariable var : varEntry.getValue()) {

				Variable artifical = new Variable(var.getName(), var.getType());

				artifical.setLowerBound(0.0d);
				artifical.setIlpLowerBound(0.0d);
				artifical.setIlpUpperBound(var.getUpperBound());
				artifical.setIlpDefaultValue(0.0d);

				artifical.setUsedInGuard(true);
				artifical.setVirtual(true);
				artifical.setVirtualVariable(var);

				variablesToWrite.get(transition).add(artifical.getName());
				hasWriteOps.put(transition, true);
				variables.put(artifical.getName(), artifical);

			}
		}

		this.variableCost = new VirtualVariableAwareWriteOperationCostWrapper(writeOperationCost, variables);

		this.finalVariableInGuard = new HashSet<Transition>();
		for (Transition t : net.getTransitions()) {
			GuardExpression guardExpression = guardExpressions.get(t);
			if (guardExpression != null && containsFinalVariables(t, guardExpression, variableCost)) {
				finalVariableInGuard.add(t);
			}
		}

		this.transitionToExpression = convertExpressions(guardExpressions);
	}

	public static SetMultimap<String, Transition> createEventClassToTransitionsMap(
			TransEvClassMapping transEvClassMapping) {
		SetMultimap<String, Transition> eventClassMap = LinkedHashMultimap.create();
		for (Entry<Transition, XEventClass> entry : transEvClassMapping.entrySet()) {
			eventClassMap.put(entry.getValue().getId(), entry.getKey());
		}
		return ImmutableSetMultimap.copyOf(eventClassMap);
	}

	private Map<Transition, IlpExpressionTemplate> convertExpressions(Map<Transition, GuardExpression> guardExpressions)
			throws DataAlignmentException {
		Builder<Transition, IlpExpressionTemplate> templatesBuilder = ImmutableMap.builder();
		int i = 0;
		for (Entry<Transition, GuardExpression> entry : guardExpressions.entrySet()) {
			try {
				Transition transition = entry.getKey();
				GuardExpression guardExpression = entry.getValue();
				String booleanVarPrefix = String.valueOf(i++).concat("_"); // unique prefix for each guard
				templatesBuilder.put(transition, new IlpExpressionTemplate(stringDiscretizer, variables, unassignedMode,
						guardExpression, booleanVarPrefix));
			} catch (NotLinearException e) {
				throw new DataAlignmentException(
						"Could not create ILP templates for a guard as one of the guard expression is not linear!", e);
			}
		}
		return templatesBuilder.build();
	}

	private static StringDiscretizer createDiscretizer(DataPetriNet net, XLog log, Map<String, String> variableMapping,
			@SuppressWarnings("rawtypes") Map<String, Class> varType, Map<String, Object> defaultValues,
			SetMultimap<String, Transition> eventIdentityToTransition,
			Map<Transition, LinkedHashSet<String>> variablesToWrite, XEventClassifier eventClassifier)
			throws StringDiscretizerException {
		return new StringDiscretizerImpl(log, getGuards(net), variableMapping, varType, defaultValues,
				eventIdentityToTransition, variablesToWrite, eventClassifier);
	}

	public static List<GuardExpression> getGuards(DataPetriNet net) {
		List<GuardExpression> guards = new ArrayList<>();
		for (Transition t : net.getTransitions()) {
			if (t instanceof PNWDTransition) {
				if (((PNWDTransition) t).getGuardExpression() != null) {
					guards.add(((PNWDTransition) t).getGuardExpression());
				}
			}
		}
		return guards;
	}

	public VariableAwareWriteOperationCost getVariableCost() {
		return variableCost;
	}

	public Map<Transition, LinkedHashSet<String>> getVariablesToWrite() {
		return variablesToWrite;
	}

	public static Map<Transition, LinkedHashSet<String>> createVariablesToWriteMap(DataPetriNet net) {
		Collection<Transition> transitions = net.getTransitions();
		Map<Transition, LinkedHashSet<String>> transitionMap = Maps.newHashMapWithExpectedSize(transitions.size());
		for (Transition transition : transitions) {
			if (transition instanceof PNWDTransition) {
				transitionMap.put(transition, createVariabletoWriteSet((PNWDTransition) transition));
			} else {
				transitionMap.put(transition, new LinkedHashSet<String>());
			}
		}
		return ImmutableMap.copyOf(transitionMap);
	}

	public static LinkedHashSet<String> createVariabletoWriteSet(PNWDTransition transition) {
		assert transition != null;
		Set<DataElement> writeOperations = transition.getWriteOperations();
		// Used a lot in iterations & We require iteration in insertion order
		LinkedHashSet<String> varToWrite = Sets.newLinkedHashSetWithExpectedSize(writeOperations.size());
		for (DataElement elem : writeOperations) {
			varToWrite.add(elem.getVarName().intern());
		}
		return varToWrite;
	}

	public static Map<Transition, String> createTransitionsToNameMap(DataPetriNet net) {
		Collection<Transition> transitions = net.getTransitions();
		Map<Transition, String> transitionMap = Maps.newHashMapWithExpectedSize(transitions.size());
		for (Transition transition : transitions) {
			transitionMap.put(transition, transition.getLabel().intern());
		}
		return ImmutableMap.copyOf(transitionMap);
	}

	public Map<Transition, String> getTransitionToName() {
		return transitionToName;
	}

	public StringDiscretizer getStringDiscretizer() {
		return stringDiscretizer;
	}

	public Map<Transition, IlpExpressionTemplate> getTransitionNameToExpression() {
		return transitionToExpression;
	}

	public Set<Transition> getMappedTransitions(XEvent e) {
		return eventIdentityToTransition.get(transEvClassMapping.getEventClassifier().getClassIdentity(e));
	}

	public boolean hasWriteOps(Transition transition) {
		return hasWriteOps.get(transition);
	}

	public SetMultimap<String, Variable> getReverseVariableMapping() {
		return reverseVariableMapping;
	}

	public XEventClassifier getEventClassifier() {
		return eventClassifier;
	}

	public SetMultimap<String, Transition> getEventIdentityToTransition() {
		return eventIdentityToTransition;
	}

	public UnassignedMode getUnassignedMode() {
		return unassignedMode;
	}

	public GuardExpression getGuardExpression(Transition transition) {
		return guardExpressions.get(transition);
	}

	public Map<String, Variable> getVariables() {
		return variables;
	}

	@SuppressWarnings("rawtypes")
	public static Map<String, Class> convertToVariableTypes(Map<String, Variable> variables) {
		return Maps.transformValues(variables, new Function<Variable, Class>() {

			public Class apply(Variable var) {
				return var.getType();
			}
		});
	}

	public boolean hasVirtualVariables() {
		return hasVirtualVariables;
	}

	public boolean finalVariablesInGuard(Transition transition) {
		return finalVariableInGuard.contains(transition);
	}

	private boolean containsFinalVariables(Transition transition, GuardExpression guardExpression,
			VariableAwareWriteOperationCost variableCost) {
		return containsFinalVariables(transition, guardExpression.getNormalVariables(), variableCost)
				|| containsFinalVariables(transition, guardExpression.getPrimeVariables(), variableCost);
	}

	private boolean containsFinalVariables(Transition transition, Collection<String> variables,
			VariableAwareWriteOperationCost variableCost) {
		for (String var : variables) {
			if (variableCost.isFinal(transition.getLabel(), getVariables().get(var))) {
				return true;
			}
		}
		return false;
	}

	public XEventClasses getEventClasses() {
		return eventClasses;
	}

	public Map<XEventClass, Integer> getEventClass2Cost() {
		return eventClass2Cost;
	}

	public Map<Transition, Integer> getTransition2Cost() {
		return transition2Cost;
	}

}
