/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.observer;

import java.text.MessageFormat;
import java.text.NumberFormat;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.logging.Level;

import org.apache.commons.math3.stat.StatUtils;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.events.Logger.MessageLevel;
import org.processmining.plugins.balancedconformance.result.BalancedDataAlignmentState;
import org.processmining.plugins.balancedconformance.result.DataAlignedTrace;
import org.processmining.plugins.balancedconformance.result.StatisticResult;

public class DataConformancePlusObserverImpl extends DataConformancePlusObserverNoOpImpl {

	private double[] dfAlignmentsCounterArray;
	private boolean[] unreliableDetected;	
	private double[] queuedStatesArray;
	private double[] cacheHitArray;
	private double[] cacheSizeArray;
	private double[] fitnessArray;
	private double[] timePerTraceArray;
	private double[] traceLength;
	private XTrace[] originalTraces;
	private long totalDataStates;
	
	private PluginContext context;
	private int progressReportWindow;
	

	public DataConformancePlusObserverImpl(final PluginContext context) {
		this.context = context;
	}

	public void finishedAlignment() {
		log("****************************");
		log("Overall statistics:");
		log(MessageFormat.format("Max Queue: {0,number,#}", StatUtils.max(queuedStatesArray)));
		log(MessageFormat.format("Avg Queue: {0,number,#.###}", StatUtils.mean(queuedStatesArray)));
		log(MessageFormat.format("Sum Data States: {0,number,#}", totalDataStates));		
		log(MessageFormat.format("Min Fitness: {0,number,#.###}", StatUtils.min(fitnessArray)));
		log(MessageFormat.format("Avg Fitness: {0,number,#.###}", StatUtils.mean(fitnessArray)));
		log(MessageFormat.format("Max Fitness: {0,number,#.###}", StatUtils.max(fitnessArray)));
		log("****************************");
	}

	@Override
	public void startAlignment(int numExpectedResults) {
		this.progressReportWindow = Math.max(numExpectedResults / 10, 10);
		this.dfAlignmentsCounterArray = new double[numExpectedResults];
		this.unreliableDetected = new boolean[numExpectedResults];
		this.fitnessArray = new double[numExpectedResults];
		this.queuedStatesArray = new double[numExpectedResults];
		this.cacheHitArray = new double[numExpectedResults];
		this.cacheSizeArray = new double[numExpectedResults];
		this.timePerTraceArray = new double[numExpectedResults];
		this.traceLength = new double[numExpectedResults];
		this.originalTraces = new XTrace[numExpectedResults];
	}

	@Override
	public void unreliableAlignmentDetected(int resultIndex, XTrace currentTrace) {
		unreliableDetected[resultIndex] = true;
		traceLength[resultIndex] = currentTrace.size();
		originalTraces[resultIndex] = currentTrace;
		timePerTraceArray[resultIndex] = -1;
	}

	@Override
	public void foundOptimalAlignment(int resultIndex, XTrace trace, DataAlignedTrace dataAlignment,
			int partialDataAlignmentsNeeded, int cacheHit, int cacheSize, long queuedStates, long dataStateCount,
			long usedTime) {
		assert dfAlignmentsCounterArray[resultIndex] == 0.0;
		dfAlignmentsCounterArray[resultIndex] = partialDataAlignmentsNeeded;
		assert queuedStatesArray[resultIndex] == 0.0;
		queuedStatesArray[resultIndex] = queuedStates;		
		assert cacheHitArray[resultIndex] == 0.0;
		cacheHitArray[resultIndex] = cacheHit;
		assert cacheSizeArray[resultIndex] == 0.0;
		cacheSizeArray[resultIndex] = cacheSize;
		assert timePerTraceArray[resultIndex] == 0.0;
		timePerTraceArray[resultIndex] = usedTime / 1000d;
		traceLength[resultIndex] = trace.size();
		originalTraces[resultIndex] = trace;
		totalDataStates = dataStateCount;
	}

	@Override
	public void calculatedFitness(int resultIndex, XTrace trace, BalancedDataAlignmentState resultState) {
		assert fitnessArray[resultIndex] == 0.0;
		fitnessArray[resultIndex] = resultState.getOverallFitness();
	}

	@Override
	public void slowDataAlignmentDetected(int resultIndex, XTrace trace, DataAlignedTrace dataAlignment,
			long usedTime) {
		String traceName = XConceptExtension.instance().extractName(trace);
		log("#" + resultIndex + " with name " + (traceName != null ? traceName : "NULL") + " took "
				+ NumberFormat.getIntegerInstance().format(usedTime / 1000d / 1000d) + " seconds to compute");
	}

	@Override
	public void processingNewTrace(int resultIndex, XTrace trace) {
		if (resultIndex % progressReportWindow == 0) {
			log("Processing Trace #" + resultIndex);
		}
	}

	@Override
	public void log(Level level, String message) {
		if (level.equals(Level.SEVERE)) {
			context.log(message, MessageLevel.ERROR);
		} else if (level.equals(Level.WARNING)) {
			context.log(message, MessageLevel.WARNING);
		} else {
			context.log(message, MessageLevel.NORMAL);
		}
	}

	public double[] getDFAlignmentsCounterArray() {
		return dfAlignmentsCounterArray;
	}

	public boolean[] getUnreliableDetected() {
		return unreliableDetected;
	}

	public double[] getFitnessArray() {
		return fitnessArray;
	}

	public void setProgressReportWindow(int progressReportWindow) {
		this.progressReportWindow = progressReportWindow;
	}

	public double[] getQueuedStatesArray() {
		return queuedStatesArray;
	}

	public double[] getTraceLength() {
		return traceLength;
	}

	public XTrace[] getOriginalTraces() {
		return originalTraces;
	}

	public Map<String, StatisticResult> getStatisticResults() {
		Map<String, StatisticResult> statistics = new LinkedHashMap<>();
		statistics.put(StatisticResult.FITNESS,
				new StatisticResult("Fitness", getFitnessArray(), "Fitness", "Frequency"));
		statistics.put(StatisticResult.MILP_PROBLEMS_SOLVED,
				new StatisticResult("# Solved MILPs", getDFAlignmentsCounterArray(), "# Solved MILPs", "Frequency"));
		statistics.put(StatisticResult.QUEUED_STATES, new StatisticResult("# Queued States", getQueuedStatesArray(),
				"# Partial Alignments Queued", "Frequency"));
		statistics.put(StatisticResult.UNRELIABLE_ALIGNMENTS,
				new StatisticResult("# Unreliable Alignments", getUnreliableDetected(), null, null));
		statistics.put(StatisticResult.CACHE_HITS,
				new StatisticResult("# Cache Hits", cacheHitArray, "Cache Hits", "Frequency"));
		statistics.put(StatisticResult.CACHE_SIZE,
				new StatisticResult("Cache Size", cacheSizeArray, "Cache Size", "Frequency"));
		statistics.put(StatisticResult.TIME_PER_TRACE,
				new StatisticResult("Computation Time (ms)", timePerTraceArray, "Time", "Frequency"));
		statistics.put(StatisticResult.TRACE_LENGTH,
				new StatisticResult("Trace Length", traceLength, "Trace Length", "Frequency"));
		return statistics;
	}

}